#!/usr/bin/env python
"""Basic test of the Calc module on 2D data."""
import datetime
from os.path import isfile
import shutil
import unittest
import pytest
import itertools

import cftime
import numpy as np
import xarray as xr

from aospy import Var
from aospy.calc import Calc, _add_metadata_as_attrs, _replace_pressure
from aospy.internal_names import ETA_STR
from aospy.utils.vertcoord import p_eta, dp_eta, p_level, dp_level
from .data.objects.examples import (
    example_proj, example_model, example_run, var_not_time_defined,
    condensation_rain, convection_rain, precip, sphum, globe, sahel, p, dp
)


@pytest.mark.filterwarnings('ignore:The enable_cftimeindex')
def _test_output_attrs(calc, dtype_out):
    with xr.set_options(enable_cftimeindex=True):
        with xr.open_dataset(calc.path_out[dtype_out]) as data:
            expected_units = calc.var.units
            if calc.dtype_out_vert == 'vert_int':
                if expected_units != '':
                    expected_units = ("(vertical integral of {0}):"
                                      " {0} m)").format(expected_units)
                else:
                    expected_units = ("(vertical integral of quantity"
                                      " with unspecified units)")
            expected_description = calc.var.description
            for name, arr in data.data_vars.items():
                assert expected_units == arr.attrs['units']
                assert expected_description == arr.attrs['description']


def _clean_test_direcs():
    for direc in [example_proj.direc_out, example_proj.tar_direc_out]:
        try:
            shutil.rmtree(direc)
        except OSError:
            pass


def _test_files_and_attrs(calc, dtype_out):
    assert isfile(calc.path_out[dtype_out])
    assert isfile(calc.path_tar_out)
    _test_output_attrs(calc, dtype_out)


_2D_DATE_RANGES = {
    'datetime': (datetime.datetime(4, 1, 1), datetime.datetime(6, 12, 31)),
    'datetime64': (np.datetime64('0004-01-01'), np.datetime64('0006-12-31')),
    'cftime': (cftime.DatetimeNoLeap(4, 1, 1),
               cftime.DatetimeNoLeap(6, 12, 31)),
    'str': ('0004', '0006')
}
_3D_DATE_RANGES = {
    'datetime': (datetime.datetime(6, 1, 1), datetime.datetime(6, 1, 31)),
    'datetime64': (np.datetime64('0006-01-01'), np.datetime64('0006-01-31')),
    'cftime': (cftime.DatetimeNoLeap(6, 1, 1),
               cftime.DatetimeNoLeap(6, 1, 31)),
    'str': ('0006', '0006')
}
_2D_VARS = {'basic': condensation_rain, 'composite': precip}
_2D_DTYPE_OUT_VERT = {'None': None}
_2D_DTYPE_IN_VERT = {'None': None}
_3D_VARS = {'3D': sphum}
_3D_DTYPE_OUT_VERT = {'vert_int': 'vert_int',
                      'vert_av': 'vert_av'}
_3D_DTYPE_IN_VERT = {'sigma': 'sigma'}
_CASES = (
    list(itertools.product(_2D_DATE_RANGES.items(), _2D_VARS.items(),
                           _2D_DTYPE_IN_VERT.items(),
                           _2D_DTYPE_OUT_VERT.items())) +
    list(itertools.product(_3D_DATE_RANGES.items(), _3D_VARS.items(),
                           _3D_DTYPE_IN_VERT.items(),
                           _3D_DTYPE_OUT_VERT.items()))
)
_CALC_TESTS = {}
for ((date_type, date_range), (test_type, var),
     (vert_in_label, vert_in), (vert_out_label, vert_out)) in _CASES:
    _CALC_TESTS['{}-{}-{}-{}'.format(
        date_type, test_type, vert_in_label, vert_out_label)] = (
            date_range, var, vert_in, vert_out)


@pytest.fixture(params=_CALC_TESTS.values(), ids=list(_CALC_TESTS.keys()))
def test_params(request):
    date_range, var, vert_in, vert_out = request.param
    yield {
        'proj': example_proj,
        'model': example_model,
        'run': example_run,
        'var': var,
        'date_range': date_range,
        'intvl_in': 'monthly',
        'dtype_in_time': 'ts',
        'dtype_in_vert': vert_in,
        'dtype_out_vert': vert_out
    }
    _clean_test_direcs()


@pytest.mark.filterwarnings('ignore:The enable_cftimeindex')
def test_annual_mean(test_params):
    calc = Calc(intvl_out='ann', dtype_out_time='av', **test_params)
    calc.compute()
    _test_files_and_attrs(calc, 'av')


@pytest.mark.filterwarnings('ignore:The enable_cftimeindex')
def test_annual_ts(test_params):
    calc = Calc(intvl_out='ann', dtype_out_time='ts', **test_params)
    calc.compute()
    _test_files_and_attrs(calc, 'ts')


@pytest.mark.filterwarnings('ignore:The enable_cftimeindex')
def test_seasonal_mean(test_params):
    calc = Calc(intvl_out='djf', dtype_out_time='av', **test_params)
    calc.compute()
    _test_files_and_attrs(calc, 'av')


@pytest.mark.filterwarnings('ignore:The enable_cftimeindex')
def test_seasonal_ts(test_params):
    calc = Calc(intvl_out='djf', dtype_out_time='ts', **test_params)
    calc.compute()
    _test_files_and_attrs(calc, 'ts')


@pytest.mark.filterwarnings('ignore:The enable_cftimeindex')
def test_monthly_mean(test_params):
    calc = Calc(intvl_out=1, dtype_out_time='av', **test_params)
    calc.compute()
    _test_files_and_attrs(calc, 'av')


@pytest.mark.filterwarnings('ignore:The enable_cftimeindex')
def test_monthly_ts(test_params):
    calc = Calc(intvl_out=1, dtype_out_time='ts', **test_params)
    calc.compute()
    _test_files_and_attrs(calc, 'ts')


@pytest.mark.filterwarnings('ignore:The enable_cftimeindex')
def test_simple_reg_av(test_params):
    calc = Calc(intvl_out='ann', dtype_out_time='reg.av', region=[globe],
                **test_params)
    calc.compute()
    _test_files_and_attrs(calc, 'reg.av')


@pytest.mark.filterwarnings('ignore:The enable_cftimeindex')
def test_simple_reg_ts(test_params):
    calc = Calc(intvl_out='ann', dtype_out_time='reg.ts', region=[globe],
                **test_params)
    calc.compute()
    _test_files_and_attrs(calc, 'reg.ts')


@pytest.mark.filterwarnings('ignore:Mean of empty slice')
@pytest.mark.filterwarnings('ignore:The enable_cftimeindex')
def test_complex_reg_av(test_params):
    calc = Calc(intvl_out='ann', dtype_out_time='reg.av', region=[sahel],
                **test_params)
    calc.compute()
    _test_files_and_attrs(calc, 'reg.av')


test_params_not_time_defined = {
    'proj': example_proj,
    'model': example_model,
    'run': example_run,
    'var': var_not_time_defined,
    'date_range': 'default',
    'intvl_in': 'monthly',
    'dtype_in_time': 'av',
    'intvl_out': 1,
}


@pytest.mark.parametrize('dtype_out_time', [None, []])
def test_calc_object_no_time_options(dtype_out_time):
    test_params_not_time_defined['dtype_out_time'] = dtype_out_time
    calc = Calc(**test_params_not_time_defined)
    if isinstance(dtype_out_time, list):
        assert calc.dtype_out_time == tuple(dtype_out_time)
    else:
        assert calc.dtype_out_time == tuple([dtype_out_time])


@pytest.mark.parametrize(
    'dtype_out_time',
    ['av', 'std', 'ts', 'reg.av', 'reg.std', 'reg.ts'])
def test_calc_object_string_time_options(dtype_out_time):
    test_params_not_time_defined['dtype_out_time'] = dtype_out_time
    with pytest.raises(ValueError):
        Calc(**test_params_not_time_defined)


def test_calc_object_time_options():
    time_options = ['av', 'std', 'ts', 'reg.av', 'reg.std', 'reg.ts']
    for i in range(1, len(time_options) + 1):
        for time_option in list(itertools.permutations(time_options, i)):
            if time_option != ('None',):
                test_params_not_time_defined['dtype_out_time'] = time_option
                with pytest.raises(ValueError):
                    Calc(**test_params_not_time_defined)


@pytest.mark.parametrize(
    ('units', 'description', 'dtype_out_vert', 'expected_units',
     'expected_description'),
    [('', '', None, '', ''),
     ('m', '', None, 'm', ''),
     ('', 'rain', None, '', 'rain'),
     ('m', 'rain', None, 'm', 'rain'),
     ('', '', 'vert_av', '', ''),
     ('m', '', 'vert_av', 'm', ''),
     ('', 'rain', 'vert_av', '', 'rain'),
     ('m', 'rain', 'vert_av', 'm', 'rain'),
     ('', '', 'vert_int',
      '(vertical integral of quantity with unspecified units)', ''),
     ('m', '', 'vert_int',
      '(vertical integral of m): m kg m^-2)', ''),
     ('', 'rain', 'vert_int',
      '(vertical integral of quantity with unspecified units)', 'rain'),
     ('m', 'rain', 'vert_int',
      '(vertical integral of m): m kg m^-2)', 'rain')])
def test_attrs(units, description, dtype_out_vert, expected_units,
               expected_description):
    da = xr.DataArray(None)
    ds = xr.Dataset({'bar': 'foo', 'boo': 'baz'})
    da = _add_metadata_as_attrs(da, units, description, dtype_out_vert)
    ds = _add_metadata_as_attrs(ds, units, description, dtype_out_vert)
    assert expected_units == da.attrs['units']
    assert expected_description == da.attrs['description']
    for name, arr in ds.data_vars.items():
        assert expected_units == arr.attrs['units']
        assert expected_description == arr.attrs['description']


@pytest.fixture()
def recursive_test_params():
    basic_params = {
        'proj': example_proj,
        'model': example_model,
        'run': example_run,
        'var': condensation_rain,
        'date_range': (datetime.datetime(4, 1, 1),
                       datetime.datetime(6, 12, 31)),
        'intvl_in': 'monthly',
        'dtype_in_time': 'ts'
    }
    recursive_params = basic_params.copy()

    recursive_condensation_rain = Var(
        name='recursive_condensation_rain',
        variables=(precip, convection_rain), func=lambda x, y: x - y,
        def_time=True)
    recursive_params['var'] = recursive_condensation_rain

    yield (basic_params, recursive_params)

    _clean_test_direcs()


# Remove autoclose argument in xarray version 0.12
@pytest.mark.filterwarnings('ignore:The autoclose')
@pytest.mark.filterwarnings('ignore:The enable_cftimeindex')
def test_recursive_calculation(recursive_test_params):
    basic_params, recursive_params = recursive_test_params

    calc = Calc(intvl_out='ann', dtype_out_time='av', **basic_params)
    calc = calc.compute()
    with xr.set_options(enable_cftimeindex=True):
        expected = xr.open_dataset(
            calc.path_out['av'], autoclose=True)['condensation_rain']
    _test_files_and_attrs(calc, 'av')

    calc = Calc(intvl_out='ann', dtype_out_time='av', **recursive_params)
    calc = calc.compute()
    with xr.set_options(enable_cftimeindex=True):
        result = xr.open_dataset(
            calc.path_out['av'], autoclose=True)['recursive_condensation_rain']
    _test_files_and_attrs(calc, 'av')

    xr.testing.assert_equal(expected, result)


def test_compute_pressure():
    calc = Calc(
        intvl_out='ann',
        dtype_out_time='av',
        var=p,
        proj=example_proj,
        model=example_model,
        run=example_run,
        date_range=('0006', '0006'),
        intvl_in='monthly',
        dtype_in_time='ts',
        dtype_in_vert='sigma',
        dtype_out_vert=None
    )
    calc.compute()
    _test_files_and_attrs(calc, 'av')
    _clean_test_direcs()


def test_compute_pressure_thicknesses():
    calc = Calc(
        intvl_out='ann',
        dtype_out_time='av',
        var=dp,
        proj=example_proj,
        model=example_model,
        run=example_run,
        date_range=('0006', '0006'),
        intvl_in='monthly',
        dtype_in_time='ts',
        dtype_in_vert='sigma',
        dtype_out_vert=None
    )
    calc.compute()
    _test_files_and_attrs(calc, 'av')
    _clean_test_direcs()


@pytest.mark.parametrize(
    ['dtype_in_vert', 'expected'],
    [(ETA_STR, [p_eta, dp_eta, condensation_rain, 5]),
     ('pressure', [p_level, dp_level, condensation_rain, 5])])
def test_replace_pressure(dtype_in_vert, expected):
    arguments = [p, dp, condensation_rain, 5]
    p_in, dp_in, cond, num = arguments
    p_expected, dp_expected, cond_expected, num_expected = expected
    assert p_in.func != p_expected.func
    assert dp_in.func != dp_expected.func
    result = _replace_pressure(arguments, dtype_in_vert)
    p_out, dp_out, cond_out, num_out = result
    assert p_out.func == p_expected.func
    assert dp_out.func == dp_expected.func
    assert cond_out.func == cond_expected.func
    assert num_out == num_expected


if __name__ == '__main__':
    unittest.main()



"""Functionality for storing attributes of a model run or obs product."""
from .utils.times import datetime_or_default


class Run(object):
    """An object that describes a single model 'run' (i.e. simulation).

    Each `Run` object is associated with a parent `Model` object.  This
    parent attribute is not set by `Run` itself, however; it is set during
    the instantation of the parent `Model` object.

    If aospy is being used to work with non climate-model data, the `Run`
    object can be used e.g. to represent different versions of a gridded
    observational data product, with the parent `Model` representing that
    data product more generally.

    Attributes
    ----------
    name : str
        The run's name
    description : str
        A description of the run
    proj : {None, aospy.Proj}
        The run's parent aospy.Proj object
    default_start_date, default_end_date : datetime.datetime
        The default start and end dates of any calculations using this Run
    data_loader : aospy.DataLoader
        The aospy.DataLoader object used to find data on disk corresponding
        to this Run object

    """

    def __init__(self, name=None, description=None, proj=None,
                 default_start_date=None,
                 default_end_date=None,
                 data_loader=None):
        """Instantiate a `Run` object.

        Parameters
        ----------
        name : str
            The run's name.  This must be unique from that of any other
            `Run` objects being used by the parent `Model`.
        description : str, optional
            A description of the model.  This is not used internally by
            aospy; it is solely for the user's information.
        proj : {None, aospy.Proj}, optional
            The parent Proj object.
        data_loader : aospy.DataLoader
            The `DataLoader` object used to find the data on disk to be used
            as inputs for aospy calculations for this run.
        default_start_date, default_end_date : datetime.datetime, optional
            Default start and end dates of calculations to be performed using
            this Model.

        See Also
        --------
        aospy.DataLoader, aospy.Model

        """
        self.name = '' if name is None else name
        self.description = '' if description is None else description
        self.proj = proj

        self.default_start_date = datetime_or_default(
            default_start_date, getattr(data_loader, 'data_start_date', None))
        self.default_end_date = datetime_or_default(
            default_end_date, getattr(data_loader, 'data_end_date', None))

        self.data_loader = data_loader

    def __str__(self):
        return 'aospy.Run instance "%s"' % self.name

    __repr__ = __str__



"""Functionality for specifying and cycling through multiple calculations."""
from __future__ import print_function

from distutils.version import LooseVersion
from multiprocessing import cpu_count

import dask
import dask.bag as db
import distributed
import itertools
import logging
import pprint
import traceback

from .calc import Calc, _TIME_DEFINED_REDUCTIONS
from .region import Region
from .var import Var


_OBJ_LIB_STR = 'library'
_PROJECTS_STR = 'projects'
_MODELS_STR = 'models'
_RUNS_STR = 'runs'
_REGIONS_STR = 'regions'
_VARIABLES_STR = 'variables'
_TAG_ATTR_MODIFIERS = dict(all='', default='default_')


class AospyException(Exception):
    """Base exception class for the aospy package."""
    pass


def _get_attr_by_tag(obj, tag, attr_name):
    """Get attribute from an object via a string tag.

    Parameters
    ----------
    obj : object from which to get the attribute
    attr_name : str
        Unmodified name of the attribute to be found.  The actual attribute
        that is returned may be modified be 'tag'.
    tag : str
        Tag specifying how to modify 'attr_name' by pre-pending it with 'tag'.
        Must be a key of the _TAG_ATTR_MODIFIERS dict.

    Returns
    -------
    the specified attribute of obj
    """
    attr_name = _TAG_ATTR_MODIFIERS[tag] + attr_name
    return getattr(obj, attr_name)


def _permuted_dicts_of_specs(specs):
    """Create {name: value} dict, one each for every permutation.

    Each permutation becomes a dictionary, with the keys being the attr names
    and the values being the corresponding value for that permutation.  These
    dicts can then be directly passed to the Calc constructor.

    """
    permuter = itertools.product(*specs.values())
    return [dict(zip(specs.keys(), perm)) for perm in permuter]


def _merge_dicts(*dict_args):
    """Merge the given dictionaries into single dict.

    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.

    From http://stackoverflow.com/a/26853961/1706640
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def _input_func_py2_py3():
    """Find function for reading user input that works on Python 2 and 3.

    See e.g. http://stackoverflow.com/questions/21731043
    """
    try:
        input = raw_input
    except NameError:
        import builtins
        input = builtins.input
    return input


def _user_verify(input_func=_input_func_py2_py3(),
                 prompt='Perform these computations? [y/n] '):
    """Prompt the user for verification."""
    if not input_func(prompt).lower()[0] == 'y':
        raise AospyException('Execution cancelled by user.')


def _get_all_objs_of_type(type_, parent):
    """Get all attributes of the given type from the given object.

    Parameters
    ----------
    type_ : The desired type
    parent : The object from which to get the attributes with type matching
        'type_'

    Returns
    -------
    A list (possibly empty) of attributes from 'parent'
    """
    return set([obj for obj in parent.__dict__.values()
                if isinstance(obj, type_)])


class CalcSuite(object):
    """Suite of Calc objects generated from provided specifications."""

    _CORE_SPEC_NAMES = {_OBJ_LIB_STR, _PROJECTS_STR, _MODELS_STR, _RUNS_STR}
    _AUX_SPEC_NAMES = {_VARIABLES_STR,
                       _REGIONS_STR,
                       'date_ranges',
                       'input_time_intervals',
                       'input_time_datatypes',
                       'input_time_offsets',
                       'input_vertical_datatypes',
                       'output_time_intervals',
                       'output_time_regional_reductions',
                       'output_vertical_reductions'}
    _NAMES_SUITE_TO_CALC = {
        _PROJECTS_STR: 'proj',
        _MODELS_STR: 'model',
        _RUNS_STR: 'run',
        _VARIABLES_STR: 'var',
        _REGIONS_STR: 'region',
        'date_ranges': 'date_range',
        'input_time_intervals': 'intvl_in',
        'input_time_datatypes': 'dtype_in_time',
        'input_time_offsets': 'time_offset',
        'input_vertical_datatypes': 'dtype_in_vert',
        'output_time_intervals': 'intvl_out',
        'output_time_regional_reductions': 'dtype_out_time',
        'output_vertical_reductions': 'dtype_out_vert',
    }

    def __init__(self, calc_suite_specs):
        self._specs_in = calc_suite_specs
        self._obj_lib = self._specs_in[_OBJ_LIB_STR]

    def _get_requested_spec(self, obj, spec_name):
        """Helper to translate user specifications to needed objects."""
        requested = self._specs_in[spec_name]
        if isinstance(requested, str):
            return _get_attr_by_tag(obj, requested, spec_name)
        else:
            return requested

    def _permute_core_specs(self):
        """Generate all requested combinations of the core objects."""
        obj_trees = []
        projects = self._get_requested_spec(self._obj_lib, _PROJECTS_STR)
        for project in projects:
            models = self._get_requested_spec(project, _MODELS_STR)
            for model in models:
                runs = self._get_requested_spec(model, _RUNS_STR)
                for run in runs:
                    obj_trees.append({
                        self._NAMES_SUITE_TO_CALC[_PROJECTS_STR]: project,
                        self._NAMES_SUITE_TO_CALC[_MODELS_STR]: model,
                        self._NAMES_SUITE_TO_CALC[_RUNS_STR]: run,
                    })
        return obj_trees

    def _get_regions(self):
        """Get the requested regions."""
        if self._specs_in[_REGIONS_STR] == 'all':
            return [_get_all_objs_of_type(
                Region, getattr(self._obj_lib, 'regions', self._obj_lib)
            )]
        else:
            return [set(self._specs_in[_REGIONS_STR])]

    def _get_variables(self):
        """Get the requested variables."""
        if self._specs_in[_VARIABLES_STR] == 'all':
            return _get_all_objs_of_type(
                Var, getattr(self._obj_lib, 'variables', self._obj_lib)
            )
        else:
            return set(self._specs_in[_VARIABLES_STR])

    def _get_date_ranges(self):
        """Parse the input to get the desired date ranges."""
        if self._specs_in['date_ranges'] == 'default':
            return ['default']
        else:
            return self._specs_in['date_ranges']

    def _get_time_reg_reducts(self):
        """Parse the input to get the desired spatiotemporal reductions."""
        return [self._specs_in['output_time_regional_reductions']]

    def _get_aux_specs(self):
        """Get and pre-process all of the non-core specifications."""
        # Drop the "core" specifications, which are handled separately.
        specs = self._specs_in.copy()
        [specs.pop(core) for core in self._CORE_SPEC_NAMES]

        specs[_REGIONS_STR] = self._get_regions()
        specs[_VARIABLES_STR] = self._get_variables()
        specs['date_ranges'] = self._get_date_ranges()
        specs['output_time_regional_reductions'] = self._get_time_reg_reducts()

        return specs

    def _permute_aux_specs(self):
        """Generate all permutations of the non-core specifications."""
        # Convert to attr names that Calc is expecting.
        calc_aux_mapping = self._NAMES_SUITE_TO_CALC.copy()
        # Special case: manually add 'library' to mapping
        calc_aux_mapping[_OBJ_LIB_STR] = None
        [calc_aux_mapping.pop(core) for core in self._CORE_SPEC_NAMES]

        specs = self._get_aux_specs()
        for suite_name, calc_name in calc_aux_mapping.items():
            specs[calc_name] = specs.pop(suite_name)
        return _permuted_dicts_of_specs(specs)

    def _combine_core_aux_specs(self):
        """Combine permutations over core and auxilliary Calc specs."""
        all_specs = []
        for core_dict in self._permute_core_specs():
            for aux_dict in self._permute_aux_specs():
                all_specs.append(_merge_dicts(core_dict, aux_dict))
        return all_specs

    def create_calcs(self):
        """Generate a Calc object for each requested parameter combination."""
        specs = self._combine_core_aux_specs()
        for spec in specs:
            spec['dtype_out_time'] = _prune_invalid_time_reductions(spec)
        return [Calc(**sp) for sp in specs]


def _prune_invalid_time_reductions(spec):
    """Prune time reductions of spec with no time dimension."""
    valid_reductions = []
    if not spec['var'].def_time and spec['dtype_out_time'] is not None:
        for reduction in spec['dtype_out_time']:
            if reduction not in _TIME_DEFINED_REDUCTIONS:
                valid_reductions.append(reduction)
            else:
                msg = ("Var {0} has no time dimension "
                       "for the given time reduction "
                       "{1} so this calculation will "
                       "be skipped".format(spec['var'].name, reduction))
                logging.info(msg)
    else:
        valid_reductions = spec['dtype_out_time']
    return valid_reductions


def _compute_or_skip_on_error(calc, compute_kwargs):
    """Execute the Calc, catching and logging exceptions, but don't re-raise.

    Prevents one failed calculation from stopping a larger requested set
    of calculations.
    """
    try:
        return calc.compute(**compute_kwargs)
    except Exception:
        msg = ("Skipping aospy calculation `{0}` due to error with the "
               "following traceback: \n{1}")
        logging.warning(msg.format(calc, traceback.format_exc()))
        return None


def _submit_calcs_on_client(calcs, client, func):
    """Submit calculations via dask.bag and a distributed client"""
    logging.info('Connected to client: {}'.format(client))
    if LooseVersion(dask.__version__) < '0.18':
        dask_option_setter = dask.set_options
    else:
        dask_option_setter = dask.config.set
    with dask_option_setter(get=client.get):
        return db.from_sequence(calcs).map(func).compute()


def _n_workers_for_local_cluster(calcs):
    """The number of workers used in a LocalCluster

    An upper bound is set at the cpu_count or the number of calcs submitted,
    depending on which is smaller.  This is to prevent more workers from
    being started than needed (but also to prevent too many workers from
    being started in the case that a large number of calcs are submitted).
    """
    return min(cpu_count(), len(calcs))


def _exec_calcs(calcs, parallelize=False, client=None, **compute_kwargs):
    """Execute the given calculations.

    Parameters
    ----------
    calcs : Sequence of ``aospy.Calc`` objects
    parallelize : bool, default False
        Whether to submit the calculations in parallel or not
    client : distributed.Client or None
        The distributed Client used if parallelize is set to True; if None
        a distributed LocalCluster is used.
    compute_kwargs : dict of keyword arguments passed to ``Calc.compute``

    Returns
    -------
    A list of the values returned by each Calc object that was executed.
    """
    if parallelize:
        def func(calc):
            """Wrap _compute_or_skip_on_error to require only the calc
            argument"""
            if 'write_to_tar' in compute_kwargs:
                compute_kwargs['write_to_tar'] = False
            return _compute_or_skip_on_error(calc, compute_kwargs)

        if client is None:
            n_workers = _n_workers_for_local_cluster(calcs)
            with distributed.LocalCluster(n_workers=n_workers) as cluster:
                with distributed.Client(cluster) as client:
                    result = _submit_calcs_on_client(calcs, client, func)
        else:
            result = _submit_calcs_on_client(calcs, client, func)
        if compute_kwargs['write_to_tar']:
            _serial_write_to_tar(calcs)
        return result
    else:
        return [_compute_or_skip_on_error(calc, compute_kwargs)
                for calc in calcs]


def _serial_write_to_tar(calcs):
    for calc in calcs:
        if calc.proj.tar_direc_out:
            for dtype_out_time in calc.dtype_out_time:
                calc._write_to_tar(dtype_out_time)


def _print_suite_summary(calc_suite_specs):
    """Print summary of requested calculations."""
    return ('\nRequested aospy calculations:\n' +
            pprint.pformat(calc_suite_specs) + '\n')


def submit_mult_calcs(calc_suite_specs, exec_options=None):
    """Generate and execute all specified computations.

    Once the calculations are prepped and submitted for execution, any
    calculation that triggers any exception or error is skipped, and the rest
    of the calculations proceed unaffected.  This prevents an error in a single
    calculation from crashing a large suite of calculations.

    Parameters
    ----------
    calc_suite_specs : dict
        The specifications describing the full set of calculations to be
        generated and potentially executed.  Accepted keys and their values:

        library : module or package comprising an aospy object library
            The aospy object library for these calculations.
        projects : list of aospy.Proj objects
            The projects to permute over.
        models : 'all', 'default', or list of aospy.Model objects
            The models to permute over.  If 'all', use all models in the
            ``models`` attribute of each ``Proj``.  If 'default', use all
            models in the ``default_models`` attribute of each ``Proj``.
        runs : 'all', 'default', or list of aospy.Run objects
            The runs to permute over.  If 'all', use all runs in the
            ``runs`` attribute of each ``Model``.  If 'default', use all
            runs in the ``default_runs`` attribute of each ``Model``.
        variables : list of aospy.Var objects
            The variables to be calculated.
        regions : 'all' or list of aospy.Region objects
            The region(s) over which any regional reductions will be performed.
            If 'all', use all regions in the ``regions`` attribute of each
            ``Proj``.
        date_ranges : 'default' or a list of tuples
            The range of dates (inclusive) over which to perform calculations.
            If 'default', use the ``default_start_date`` and
            ``default_end_date`` attribute of each ``Run``. Else provide a
            list of tuples, each containing a pair of start and end dates,
            such as  ``date_ranges=[(start, end)]`` where ``start`` and
            ``end`` are each ``datetime.datetime`` objects, partial
            datetime strings (e.g. '0001'), ``np.datetime64`` objects, or
            ``cftime.datetime`` objects.

        output_time_intervals : {'ann', season-string, month-integer}
            The sub-annual time interval over which to aggregate.

            - 'ann' : Annual mean
            - season-string : E.g. 'JJA' for June-July-August
            - month-integer : 1 for January, 2 for February, etc.  Each one is
                  a separate reduction, e.g. [1, 2] would produce averages (or
                  other specified time reduction) over all Januaries, and
                  separately over all Februaries.

        output_time_regional_reductions : list of reduction string identifiers
            Unlike most other keys, these are not permuted over when creating
            the :py:class:`aospy.Calc` objects that execute the calculations;
            each :py:class:`aospy.Calc` performs all of the specified
            reductions.  Accepted string identifiers are:

            - Gridpoint-by-gridpoint output:

              - 'av' : Gridpoint-by-gridpoint time-average
              - 'std' : Gridpoint-by-gridpoint temporal standard deviation
              - 'ts' : Gridpoint-by-gridpoint time-series

            - Averages over each region specified via `region`:

              - 'reg.av', 'reg.std', 'reg.ts' : analogous to 'av', 'std', 'ts'

        output_vertical_reductions : {None, 'vert_av', 'vert_int'}, optional
            How to reduce the data vertically:

            - None : no vertical reduction
            - 'vert_av' : mass-weighted vertical average
            - 'vert_int' : mass-weighted vertical integral
        input_time_intervals : {'annual', 'monthly', 'daily', '#hr'}
            A string specifying the time resolution of the input data.  In
            '#hr' above, the '#' stands for a number, e.g. 3hr or 6hr, for
            sub-daily output.  These are the suggested specifiers, but others
            may be used if they are also used by the DataLoaders for the given
            Runs.
        input_time_datatypes : {'inst', 'ts', 'av'}
            What the time axis of the input data represents:

            - 'inst' : Timeseries of instantaneous values
            - 'ts' : Timeseries of averages over the period of each time-index
            - 'av' : A single value averaged over a date range

        input_vertical_datatypes : {False, 'pressure', 'sigma'}, optional
            The vertical coordinate system used by the input data:

            - False : not defined vertically
            - 'pressure' : pressure coordinates
            - 'sigma' : hybrid sigma-pressure coordinates

        input_time_offsets : {None, dict}, optional
            How to offset input data in time to correct for metadata errors

            - None : no time offset applied
            - dict : e.g. ``{'hours': -3}`` to offset times by -3 hours
              See :py:meth:`aospy.utils.times.apply_time_offset`.

    exec_options : dict or None (default None)
        Options regarding how the calculations are reported, submitted, and
        saved.  If None, default settings are used for all options.  Currently
        supported options (each should be either `True` or `False`):

        - prompt_verify : (default False) If True, print summary of
              calculations to be performed and prompt user to confirm before
              submitting for execution.
        - parallelize : (default False) If True, submit calculations in
              parallel.
        - client : distributed.Client or None (default None) The
              dask.distributed Client used to schedule computations.  If None
              and parallelize is True, a LocalCluster will be started.
        - write_to_tar : (default True) If True, write results of calculations
              to .tar files, one for each :py:class:`aospy.Run` object.
              These tar files have an identical directory structures the
              standard output relative to their root directory, which is
              specified via the `tar_direc_out` argument of each Proj
              object's instantiation.

    Returns
    -------
    A list of the return values from each :py:meth:`aospy.Calc.compute` call
        If a calculation ran without error, this value is the
        :py:class:`aospy.Calc` object itself, with the results of its
        calculations saved in its ``data_out`` attribute.  ``data_out`` is a
        dictionary, with the keys being the temporal-regional reduction
        identifiers (e.g. 'reg.av'), and the values being the corresponding
        result.

    If any error occurred during a calculation, the return value is None.

    Raises
    ------
    AospyException
        If the ``prompt_verify`` option is set to True and the user does not
        respond affirmatively to the prompt.

    """
    if exec_options is None:
        exec_options = dict()
    if exec_options.pop('prompt_verify', False):
        print(_print_suite_summary(calc_suite_specs))
        _user_verify()
    calc_suite = CalcSuite(calc_suite_specs)
    calcs = calc_suite.create_calcs()
    if not calcs:
        raise AospyException(
            "The specified combination of parameters yielded zero "
            "calculations.  Most likely, one of the parameters is "
            "inadvertently empty."
        )
    return _exec_calcs(calcs, **exec_options)



#!/usr/bin/env python
"""Test suite for aospy.timedate module."""
import datetime
import warnings

import cftime
import numpy as np
import pandas as pd
import pytest
import xarray as xr

from itertools import product

from aospy.data_loader import set_grid_attrs_as_coords
from aospy.internal_names import (
    TIME_STR, TIME_BOUNDS_STR, BOUNDS_STR, TIME_WEIGHTS_STR,
    RAW_START_DATE_STR, RAW_END_DATE_STR, SUBSET_START_DATE_STR,
    SUBSET_END_DATE_STR
)
from aospy.automate import _merge_dicts
from aospy.utils.times import (
    apply_time_offset,
    average_time_bounds,
    monthly_mean_ts,
    monthly_mean_at_each_ind,
    ensure_datetime,
    datetime_or_default,
    month_indices,
    _month_conditional,
    extract_months,
    ensure_time_avg_has_cf_metadata,
    _assert_has_data_for_time,
    add_uniform_time_weights,
    assert_matching_time_coord,
    ensure_time_as_index,
    sel_time,
    yearly_average,
    infer_year,
    maybe_convert_to_index_date_type
)


_INVALID_DATE_OBJECTS = [1985, True, None]


def test_apply_time_offset():
    start = datetime.datetime(1900, 5, 10)
    years, months, days, hours = -2, 1, 7, 3
    # test lengths 0, 1, and >1 of input time array
    for periods in range(3):
        times = pd.date_range(start=start, freq='M', periods=periods)
        actual = apply_time_offset(xr.DataArray(times), years=years,
                                   months=months, days=days, hours=hours)
        desired = (times + pd.tseries.offsets.DateOffset(
            years=years, months=months, days=days, hours=hours
        ))
        assert actual.identical(desired)


def test_monthly_mean_ts_single_month():
    time = pd.date_range('2000-01-01', freq='6H', periods=4 * 31)
    arr = xr.DataArray(np.random.random(time.shape), dims=[TIME_STR],
                       coords={TIME_STR: time})
    desired = arr.mean(TIME_STR)
    actual = monthly_mean_ts(arr)
    np.testing.assert_allclose(actual, desired)


def test_monthly_mean_ts_submonthly():
    time = pd.date_range('2000-01-01', freq='1D', periods=365 * 3)
    arr = xr.DataArray(np.random.random(time.shape), dims=[TIME_STR],
                       coords={TIME_STR: time})
    desired = arr.resample(**{TIME_STR: '1M'}).mean(TIME_STR)
    actual = monthly_mean_ts(arr)
    assert desired.identical(actual)


def test_monthly_mean_ts_monthly():
    time = pd.date_range('2000-01-01', freq='1M', periods=120)
    arr = xr.DataArray(np.random.random(time.shape), dims=[TIME_STR],
                       coords={TIME_STR: time})
    actual = monthly_mean_ts(arr)
    assert arr.identical(actual)


@pytest.mark.filterwarnings('ignore:Mean of empty slice')
def test_monthly_mean_ts_na():
    time = pd.to_datetime(['2000-06-01', '2001-06-01'])
    arr = xr.DataArray(np.random.random(time.shape), dims=[TIME_STR],
                       coords={TIME_STR: time})
    arr = arr.resample(**{TIME_STR: '1M'}).mean(TIME_STR)
    actual = monthly_mean_ts(arr)
    desired = arr.dropna(TIME_STR)
    assert desired.identical(actual)


def test_monthly_mean_at_each_ind():
    times_submonthly = pd.to_datetime(['2000-06-01', '2000-06-15',
                                       '2000-07-04', '2000-07-19'])
    times_means = pd.to_datetime(['2000-06-01', '2000-07-01'])
    len_other_dim = 2
    arr_submonthly = xr.DataArray(
        np.random.random((len(times_submonthly), len_other_dim)),
        dims=[TIME_STR, 'dim0'], coords={TIME_STR: times_submonthly}
    )
    arr_means = xr.DataArray(
        np.random.random((len(times_means), len_other_dim)),
        dims=arr_submonthly.dims, coords={TIME_STR: times_means}
    )
    actual = monthly_mean_at_each_ind(arr_means, arr_submonthly)
    desired_values = np.stack([arr_means.values[0]] * len_other_dim +
                              [arr_means.values[1]] * len_other_dim,
                              axis=0)
    desired = xr.DataArray(desired_values, dims=arr_submonthly.dims,
                           coords=arr_submonthly.coords)
    assert actual.identical(desired)


@pytest.mark.parametrize('date', [np.datetime64('2000-01-01'),
                                  cftime.DatetimeNoLeap(1, 1, 1),
                                  datetime.datetime(1, 1, 1),
                                  '2000-01-01'])
def test_ensure_datetime_valid_input(date):
    assert ensure_datetime(date) == date


def test_ensure_datetime_invalid_input():
    with pytest.raises(TypeError):
        for obj in _INVALID_DATE_OBJECTS:
            ensure_datetime(obj)


def test_datetime_or_default():
    date = np.datetime64('2000-01-01')
    assert datetime_or_default(None, 'dummy') == 'dummy'
    assert datetime_or_default(date, 'dummy') == ensure_datetime(date)


def test_month_indices():
    np.testing.assert_array_equal(month_indices('ann'), range(1, 13))
    np.testing.assert_array_equal(month_indices('jja'),
                                  np.array([6, 7, 8]))
    with pytest.raises(ValueError):
        month_indices('dfm')
        month_indices('j')
        month_indices('q')
    assert month_indices('s') == [9]
    np.testing.assert_array_equal(month_indices('djf'),
                                  np.array([12, 1, 2]))
    assert month_indices(12) == [12]


def test_month_conditional():
    test = pd.date_range('2000-01-01', '2000-03-01', freq='M')
    test = xr.DataArray(test, dims=[TIME_STR], coords=[test])
    result_jan = _month_conditional(test, [1])
    np.testing.assert_array_equal(result_jan, np.array([True, False]))

    result_jan_feb = _month_conditional(test, [1, 2])
    np.testing.assert_array_equal(result_jan_feb, np.array([True, True]))

    result_march = _month_conditional(test, [3])
    np.testing.assert_array_equal(result_march, np.array([False, False]))

    test = pd.date_range('1999-12-31 18:00:00', '2000-01-01 00:00:00',
                         freq='6H')
    test = xr.DataArray(test, dims=[TIME_STR])
    result_jan = _month_conditional(test, [1])
    np.testing.assert_array_equal(result_jan,
                                  np.array([False, True]))

    result_jd = _month_conditional(test, [1, 12])
    np.testing.assert_array_equal(result_jd,
                                  np.array([True, True]))

    # Test month not in range
    result_march = _month_conditional(test, [3])
    np.testing.assert_array_equal(result_march,
                                  np.array([False, False]))


def test_extract_months():
    time = xr.DataArray(pd.date_range(start='2001-02-18', end='2002-07-12',
                                      freq='1D'), dims=[TIME_STR])
    months = 'mam'  # March-April-May
    desired = xr.concat([
        xr.DataArray(pd.date_range(start='2001-03-01', end='2001-05-31',
                                   freq='1D'), dims=[TIME_STR]),
        xr.DataArray(pd.date_range(start='2002-03-01', end='2002-05-31',
                                   freq='1D'), dims=[TIME_STR])
    ], dim=TIME_STR)
    actual = extract_months(time, months)
    xr.testing.assert_identical(actual, desired)


def test_extract_months_single_month():
    time = xr.DataArray(pd.date_range(start='1678-01-01', end='1678-01-31',
                                      freq='1M'), dims=[TIME_STR])
    months = 1
    desired = time
    actual = extract_months(time, months)
    xr.testing.assert_identical(actual, desired)


@pytest.fixture
def ds_time_encoded_cf():
    time_bounds = np.array([[0, 31], [31, 59], [59, 90]])
    nv = np.array([0, 1])
    time = np.array([15, 46, 74])
    data = np.zeros((3))
    ds = xr.DataArray(data,
                      coords=[time],
                      dims=[TIME_STR],
                      name='a').to_dataset()
    ds[TIME_BOUNDS_STR] = xr.DataArray(time_bounds,
                                       coords=[time, nv],
                                       dims=[TIME_STR, BOUNDS_STR],
                                       name=TIME_BOUNDS_STR)
    units_str = 'days since 2000-01-01 00:00:00'
    cal_str = 'noleap'
    ds[TIME_STR].attrs['units'] = units_str
    ds[TIME_STR].attrs['calendar'] = cal_str
    return ds


def test_ensure_time_avg_has_cf_metadata(ds_time_encoded_cf):
    ds = ds_time_encoded_cf
    time = ds[TIME_STR].values
    time_bounds = ds[TIME_BOUNDS_STR].values
    units_str = ds[TIME_STR].attrs['units']
    cal_str = ds[TIME_STR].attrs['calendar']

    with pytest.raises(KeyError):
        ds[TIME_BOUNDS_STR].attrs['units']
    with pytest.raises(KeyError):
        ds[TIME_BOUNDS_STR].attrs['calendar']

    ds = ensure_time_avg_has_cf_metadata(ds)

    result = ds[TIME_BOUNDS_STR].attrs['units']
    assert result == units_str
    result = ds[TIME_BOUNDS_STR].attrs['calendar']
    assert result == cal_str

    avg_DT_data = np.diff(time_bounds, axis=1).squeeze()
    average_DT_expected = xr.DataArray(avg_DT_data,
                                       coords=[time],
                                       dims=[TIME_STR],
                                       name=TIME_WEIGHTS_STR)
    average_DT_expected[TIME_STR].attrs['units'] = units_str
    average_DT_expected.attrs['units'] = 'days'
    average_DT_expected[TIME_STR].attrs['calendar'] = cal_str
    assert ds[TIME_WEIGHTS_STR].identical(average_DT_expected)

    assert ds[RAW_START_DATE_STR].values == [0]
    assert ds[RAW_START_DATE_STR].attrs['units'] == units_str
    assert ds[RAW_START_DATE_STR].attrs['calendar'] == cal_str

    assert ds[RAW_END_DATE_STR].values == [90]
    assert ds[RAW_END_DATE_STR].attrs['units'] == units_str
    assert ds[RAW_END_DATE_STR].attrs['calendar'] == cal_str


def test_add_uniform_time_weights():
    time = np.array([15, 46, 74])
    data = np.zeros((3))
    ds = xr.DataArray(data,
                      coords=[time],
                      dims=[TIME_STR],
                      name='a').to_dataset()
    units_str = 'days since 2000-01-01 00:00:00'
    cal_str = 'noleap'
    ds[TIME_STR].attrs['units'] = units_str
    ds[TIME_STR].attrs['calendar'] = cal_str

    with pytest.raises(KeyError):
        ds[TIME_WEIGHTS_STR]

    ds = add_uniform_time_weights(ds)
    time_weights_expected = xr.DataArray(
        [1, 1, 1], coords=ds[TIME_STR].coords, name=TIME_WEIGHTS_STR)
    time_weights_expected.attrs['units'] = 'days'
    assert ds[TIME_WEIGHTS_STR].identical(time_weights_expected)


def test_assert_has_data_for_time():
    time_bounds = np.array([[0, 31], [31, 59], [59, 90]])
    nv = np.array([0, 1])
    time = np.array([15, 46, 74])
    data = np.zeros((3))
    var_name = 'a'
    ds = xr.DataArray(data,
                      coords=[time],
                      dims=[TIME_STR],
                      name=var_name).to_dataset()
    ds[TIME_BOUNDS_STR] = xr.DataArray(time_bounds,
                                       coords=[time, nv],
                                       dims=[TIME_STR, BOUNDS_STR],
                                       name=TIME_BOUNDS_STR)
    units_str = 'days since 2000-01-01 00:00:00'
    ds[TIME_STR].attrs['units'] = units_str
    ds = ensure_time_avg_has_cf_metadata(ds)
    ds = set_grid_attrs_as_coords(ds)
    ds = xr.decode_cf(ds)
    da = ds[var_name]

    start_date = np.datetime64('2000-01-01')
    end_date = np.datetime64('2000-03-31')
    _assert_has_data_for_time(da, start_date, end_date)

    start_date_bad = np.datetime64('1999-12-31')
    end_date_bad = np.datetime64('2000-04-01')

    with pytest.raises(AssertionError):
        _assert_has_data_for_time(da, start_date_bad, end_date)

    with pytest.raises(AssertionError):
        _assert_has_data_for_time(da, start_date, end_date_bad)

    with pytest.raises(AssertionError):
        _assert_has_data_for_time(da, start_date_bad, end_date_bad)


_CFTIME_DATE_TYPES = {
    'noleap': cftime.DatetimeNoLeap,
    '365_day': cftime.DatetimeNoLeap,
    '360_day': cftime.Datetime360Day,
    'julian': cftime.DatetimeJulian,
    'all_leap': cftime.DatetimeAllLeap,
    '366_day': cftime.DatetimeAllLeap,
    'gregorian': cftime.DatetimeGregorian,
    'proleptic_gregorian': cftime.DatetimeProlepticGregorian
}


@pytest.mark.filterwarnings('ignore:The enable_cftimeindex')
@pytest.mark.filterwarnings('ignore:Unable to decode')
@pytest.mark.parametrize(['calendar', 'date_type'],
                         list(_CFTIME_DATE_TYPES.items()))
def test_assert_has_data_for_time_cftime_datetimes(calendar, date_type):
    time_bounds = np.array([[0, 2], [2, 4], [4, 6]])
    nv = np.array([0, 1])
    time = np.array([1, 3, 5])
    data = np.zeros((3))
    var_name = 'a'
    ds = xr.DataArray(data,
                      coords=[time],
                      dims=[TIME_STR],
                      name=var_name).to_dataset()
    ds[TIME_BOUNDS_STR] = xr.DataArray(time_bounds,
                                       coords=[time, nv],
                                       dims=[TIME_STR, BOUNDS_STR],
                                       name=TIME_BOUNDS_STR)
    units_str = 'days since 0002-01-02 00:00:00'
    ds[TIME_STR].attrs['units'] = units_str
    ds[TIME_STR].attrs['calendar'] = calendar
    ds = ensure_time_avg_has_cf_metadata(ds)
    ds = set_grid_attrs_as_coords(ds)

    with warnings.catch_warnings(record=True):
        with xr.set_options(enable_cftimeindex=True):
            ds = xr.decode_cf(ds)
    da = ds[var_name]

    start_date = date_type(2, 1, 2)
    end_date = date_type(2, 1, 8)

    _assert_has_data_for_time(da, start_date, end_date)

    start_date_bad = date_type(2, 1, 1)
    end_date_bad = date_type(2, 1, 9)

    with pytest.raises(AssertionError):
        _assert_has_data_for_time(da, start_date_bad, end_date)

    with pytest.raises(AssertionError):
        _assert_has_data_for_time(da, start_date, end_date_bad)

    with pytest.raises(AssertionError):
        _assert_has_data_for_time(da, start_date_bad, end_date_bad)


def test_assert_has_data_for_time_str_input():
    time_bounds = np.array([[0, 31], [31, 59], [59, 90]])
    nv = np.array([0, 1])
    time = np.array([15, 46, 74])
    data = np.zeros((3))
    var_name = 'a'
    ds = xr.DataArray(data,
                      coords=[time],
                      dims=[TIME_STR],
                      name=var_name).to_dataset()
    ds[TIME_BOUNDS_STR] = xr.DataArray(time_bounds,
                                       coords=[time, nv],
                                       dims=[TIME_STR, BOUNDS_STR],
                                       name=TIME_BOUNDS_STR)
    units_str = 'days since 2000-01-01 00:00:00'
    ds[TIME_STR].attrs['units'] = units_str
    ds = ensure_time_avg_has_cf_metadata(ds)
    ds = set_grid_attrs_as_coords(ds)
    ds = xr.decode_cf(ds)
    da = ds[var_name]

    start_date = '2000-01-01'
    end_date = '2000-03-31'
    _assert_has_data_for_time(da, start_date, end_date)

    start_date_bad = '1999-12-31'
    end_date_bad = '2000-04-01'

    # With strings these checks are disabled
    _assert_has_data_for_time(da, start_date_bad, end_date)
    _assert_has_data_for_time(da, start_date, end_date_bad)
    _assert_has_data_for_time(da, start_date_bad, end_date_bad)


def test_assert_matching_time_coord():
    rng = pd.date_range('2000-01-01', '2001-01-01', freq='M')
    arr1 = xr.DataArray(rng, coords=[rng], dims=[TIME_STR])
    arr2 = xr.DataArray(rng, coords=[rng], dims=[TIME_STR])
    assert_matching_time_coord(arr1, arr2)

    arr2 = arr2.sel(**{TIME_STR: slice('2000-03', '2000-05')})
    with pytest.raises(ValueError):
        assert_matching_time_coord(arr1, arr2)


def test_ensure_time_as_index_no_change():
    # Already properly indexed, so shouldn't be modified.
    arr = xr.DataArray([-23, 42.4], coords=[[1, 2]], dims=[TIME_STR])
    arr[TIME_STR].attrs['units'] = 'days since 2000-01-01 00:00:00'
    arr[TIME_STR].attrs['calendar'] = 'standard'
    ds = arr.to_dataset(name='a')
    ds.coords[TIME_WEIGHTS_STR] = xr.DataArray(
        [1, 1], dims=[TIME_STR], coords={TIME_STR: arr[TIME_STR]}
    )
    ds.coords[TIME_BOUNDS_STR] = xr.DataArray(
        [[0.5, 1.5], [1.5, 2.5]], dims=[TIME_STR, BOUNDS_STR],
        coords={TIME_STR: arr[TIME_STR]}
    )
    xr.testing.assert_identical(ds, ensure_time_as_index(ds))


def test_ensure_time_as_index_with_change():
    # Time bounds array doesn't index time initially, which gets fixed.
    arr = xr.DataArray([-93], dims=[TIME_STR], coords={TIME_STR: [3]})
    arr[TIME_STR].attrs['units'] = 'days since 2000-01-01 00:00:00'
    arr[TIME_STR].attrs['calendar'] = 'standard'
    ds = arr.to_dataset(name='a')
    ds.coords[TIME_WEIGHTS_STR] = xr.DataArray(
        [1], dims=[TIME_STR], coords={TIME_STR: arr[TIME_STR]}
    )
    ds.coords[TIME_BOUNDS_STR] = xr.DataArray(
        [[3.5, 4.5]], dims=[TIME_STR, BOUNDS_STR],
        coords={TIME_STR: arr[TIME_STR]}
    )
    ds = ds.isel(**{TIME_STR: 0}).expand_dims(TIME_STR)
    actual = ensure_time_as_index(ds)
    expected = arr.to_dataset(name='a')
    expected.coords[TIME_WEIGHTS_STR] = xr.DataArray(
        [1], dims=[TIME_STR], coords={TIME_STR: arr[TIME_STR]}
    )
    expected.coords[TIME_BOUNDS_STR] = xr.DataArray(
        [[3.5, 4.5]], dims=[TIME_STR, BOUNDS_STR],
        coords={TIME_STR: arr[TIME_STR]}
        )
    xr.testing.assert_identical(actual, expected)


def test_sel_time():
    time_bounds = np.array([[0, 31], [31, 59], [59, 90]])
    nv = np.array([0, 1])
    time = np.array([15, 46, 74])
    data = np.zeros((3))
    var_name = 'a'
    ds = xr.DataArray(data,
                      coords=[time],
                      dims=[TIME_STR],
                      name=var_name).to_dataset()
    ds[TIME_BOUNDS_STR] = xr.DataArray(time_bounds,
                                       coords=[time, nv],
                                       dims=[TIME_STR, BOUNDS_STR],
                                       name=TIME_BOUNDS_STR)
    units_str = 'days since 2000-01-01 00:00:00'
    ds[TIME_STR].attrs['units'] = units_str
    ds = ensure_time_avg_has_cf_metadata(ds)
    ds = set_grid_attrs_as_coords(ds)
    ds = xr.decode_cf(ds)
    da = ds[var_name]

    start_date = np.datetime64('2000-02-01')
    end_date = np.datetime64('2000-03-31')
    result = sel_time(da, start_date, end_date)
    assert result[SUBSET_START_DATE_STR].values == start_date
    assert result[SUBSET_END_DATE_STR].values == end_date


def test_yearly_average_no_mask():
    times = pd.to_datetime(['2000-06-01', '2000-06-15',
                            '2001-07-04', '2001-10-01', '2001-12-31',
                            '2004-01-01'])
    arr = xr.DataArray(np.random.random((len(times),)),
                       dims=[TIME_STR], coords={TIME_STR: times})
    dt = arr.copy(deep=True)
    dt.values = np.random.random((len(times),))

    actual = yearly_average(arr, dt)

    yr2000 = (arr[0]*dt[0] + arr[1]*dt[1]) / (dt[0] + dt[1])
    yr2001 = ((arr[2]*dt[2] + arr[3]*dt[3] + arr[4]*dt[4]) /
              (dt[2] + dt[3] + dt[4]))
    yr2004 = arr[-1]
    yrs_coord = [2000, 2001, 2004]
    yr_avgs = np.array([yr2000, yr2001, yr2004])
    desired = xr.DataArray(yr_avgs, dims=['year'], coords={'year': yrs_coord})
    xr.testing.assert_allclose(actual, desired)


def test_yearly_average_masked_data():
    times = pd.to_datetime(['2000-06-01', '2000-06-15',
                            '2001-07-04', '2001-10-01', '2001-12-31',
                            '2004-01-01'])
    arr = xr.DataArray(np.random.random((len(times),)),
                       dims=[TIME_STR], coords={TIME_STR: times})
    arr[0] = -999
    arr = arr.where(arr != -999)
    dt = arr.copy(deep=True)
    dt.values = np.random.random((len(times),))

    actual = yearly_average(arr, dt)

    yr2000 = arr[1]
    yr2001 = ((arr[2]*dt[2] + arr[3]*dt[3] + arr[4]*dt[4]) /
              (dt[2] + dt[3] + dt[4]))
    yr2004 = arr[-1]
    yrs_coord = [2000, 2001, 2004]
    yr_avgs = np.array([yr2000, yr2001, yr2004])
    desired = xr.DataArray(yr_avgs, dims=['year'], coords={'year': yrs_coord})
    xr.testing.assert_allclose(actual, desired)


def test_average_time_bounds(ds_time_encoded_cf):
    ds = ds_time_encoded_cf
    actual = average_time_bounds(ds)[TIME_STR]

    desired_values = ds[TIME_BOUNDS_STR].mean(dim=BOUNDS_STR).values
    desired = xr.DataArray(desired_values, dims=[TIME_STR],
                           coords={TIME_STR: desired_values}, name=TIME_STR)

    xr.testing.assert_identical(actual, desired)


_INFER_YEAR_TESTS = [
    (np.datetime64('2000-01-01'), 2000),
    (datetime.datetime(2000, 1, 1), 2000),
    ('2000', 2000),
    ('2000-01', 2000),
    ('2000-01-01', 2000)
]
_INFER_YEAR_TESTS = _INFER_YEAR_TESTS + [
    (date_type(2000, 1, 1), 2000) for date_type in _CFTIME_DATE_TYPES.values()]


@pytest.mark.parametrize(
    ['date', 'expected'],
    _INFER_YEAR_TESTS)
def test_infer_year(date, expected):
    assert infer_year(date) == expected


@pytest.mark.parametrize('date', ['-0001', 'A001', '01'])
def test_infer_year_invalid(date):
    with pytest.raises(ValueError):
        infer_year(date)


_DATETIME_INDEX = pd.date_range('2000-01-01', freq='M', periods=1)
_DATETIME_CONVERT_TESTS = {}
for date_label, date_type in _CFTIME_DATE_TYPES.items():
    key = 'DatetimeIndex-{}'.format(date_label)
    _DATETIME_CONVERT_TESTS[key] = (_DATETIME_INDEX, date_type(2000, 1, 1),
                                    np.datetime64('2000-01'))
_NON_CFTIME_DATES = {
    'datetime.datetime': datetime.datetime(2000, 1, 1),
    'np.datetime64': np.datetime64('2000-01-01'),
    'str': '2000'
}
for date_label, date in _NON_CFTIME_DATES.items():
    key = 'DatetimeIndex-{}'.format(date_label)
    if isinstance(date, str):
        _DATETIME_CONVERT_TESTS[key] = (_DATETIME_INDEX, date, date)
    else:
        _DATETIME_CONVERT_TESTS[key] = (_DATETIME_INDEX, date,
                                        np.datetime64('2000-01'))

_CFTIME_INDEXES = {
    'CFTimeIndex[{}]'.format(key): xr.CFTimeIndex([value(1, 1, 1)]) for
    key, value in _CFTIME_DATE_TYPES.items()
}
_CFTIME_CONVERT_TESTS = {}
for ((index_label, index),
     (date_label, date_type)) in product(_CFTIME_INDEXES.items(),
                                         _CFTIME_DATE_TYPES.items()):
    key = '{}-{}'.format(index_label, date_label)
    _CFTIME_CONVERT_TESTS[key] = (index, date_type(1, 1, 1),
                                  index.date_type(1, 1, 1))
_NON_CFTIME_DATES_0001 = {
    'datetime.datetime': datetime.datetime(1, 1, 1),
    'np.datetime64': np.datetime64('0001-01-01'),
    'str': '0001'
}
for ((idx_label, index),
     (date_label, date)) in product(_CFTIME_INDEXES.items(),
                                    _NON_CFTIME_DATES_0001.items()):
    key = '{}-{}'.format(index_label, date_label)
    if isinstance(date, str):
        _CFTIME_CONVERT_TESTS[key] = (index, date, date)
    else:
        _CFTIME_CONVERT_TESTS[key] = (index, date, index.date_type(1, 1, 1))

_CONVERT_DATE_TYPE_TESTS = _merge_dicts(_DATETIME_CONVERT_TESTS,
                                        _CFTIME_CONVERT_TESTS)


@pytest.mark.parametrize(['index', 'date', 'expected'],
                         list(_CONVERT_DATE_TYPE_TESTS.values()),
                         ids=list(_CONVERT_DATE_TYPE_TESTS.keys()))
def test_maybe_convert_to_index_date_type(index, date, expected):
    result = maybe_convert_to_index_date_type(index, date)
    assert result == expected



"""Utility functions for dealing with vertical coordinates."""
import logging

import numpy as np
import xarray as xr

from .._constants import GRAV_EARTH
from ..var import Var
from .. import internal_names


def to_radians(arr, is_delta=False):
    """Force data with units either degrees or radians to be radians."""
    # Infer the units from embedded metadata, if it's there.
    try:
        units = arr.units
    except AttributeError:
        pass
    else:
        if units.lower().startswith('degrees'):
            warn_msg = ("Conversion applied: degrees -> radians to array: "
                        "{}".format(arr))
            logging.debug(warn_msg)
            return np.deg2rad(arr)
    # Otherwise, assume degrees if the values are sufficiently large.
    threshold = 0.1*np.pi if is_delta else 4*np.pi
    if np.max(np.abs(arr)) > threshold:
        warn_msg = ("Conversion applied: degrees -> radians to array: "
                    "{}".format(arr))
        logging.debug(warn_msg)
        return np.deg2rad(arr)
    return arr


def to_pascal(arr, is_dp=False):
    """Force data with units either hPa or Pa to be in Pa."""
    threshold = 400 if is_dp else 1200
    if np.max(np.abs(arr)) < threshold:
        warn_msg = "Conversion applied: hPa -> Pa to array: {}".format(arr)
        logging.debug(warn_msg)
        return arr*100.
    return arr


def to_hpa(arr):
    """Convert pressure array from Pa to hPa (if needed)."""
    if np.max(np.abs(arr)) > 1200.:
        warn_msg = "Conversion applied: Pa -> hPa to array: {}".format(arr)
        logging.debug(warn_msg)
        return arr / 100.
    return arr


def phalf_from_ps(bk, pk, ps):
    """Compute pressure of half levels of hybrid sigma-pressure coordinates."""
    return ps*bk + pk


def replace_coord(arr, old_dim, new_dim, new_coord):
    """Replace a coordinate with new one; new and old must have same shape."""
    new_arr = arr.rename({old_dim: new_dim})
    new_arr[new_dim] = new_coord
    return new_arr


def to_pfull_from_phalf(arr, pfull_coord):
    """Compute data at full pressure levels from values at half levels."""
    phalf_top = arr.isel(**{internal_names.PHALF_STR: slice(1, None)})
    phalf_top = replace_coord(phalf_top, internal_names.PHALF_STR,
                              internal_names.PFULL_STR, pfull_coord)

    phalf_bot = arr.isel(**{internal_names.PHALF_STR: slice(None, -1)})
    phalf_bot = replace_coord(phalf_bot, internal_names.PHALF_STR,
                              internal_names.PFULL_STR, pfull_coord)
    return 0.5*(phalf_bot + phalf_top)


def to_phalf_from_pfull(arr, val_toa=0, val_sfc=0):
    """Compute data at half pressure levels from values at full levels.

    Could be the pressure array itself, but it could also be any other data
    defined at pressure levels.  Requires specification of values at surface
    and top of atmosphere.
    """
    phalf = np.zeros((arr.shape[0] + 1, arr.shape[1], arr.shape[2]))
    phalf[0] = val_toa
    phalf[-1] = val_sfc
    phalf[1:-1] = 0.5*(arr[:-1] + arr[1:])
    return phalf


def pfull_from_ps(bk, pk, ps, pfull_coord):
    """Compute pressure at full levels from surface pressure."""
    return to_pfull_from_phalf(phalf_from_ps(bk, pk, ps), pfull_coord)


def d_deta_from_phalf(arr, pfull_coord):
    """Compute pressure level thickness from half level pressures."""
    d_deta = arr.diff(dim=internal_names.PHALF_STR, n=1)
    return replace_coord(d_deta, internal_names.PHALF_STR,
                         internal_names.PFULL_STR, pfull_coord)


def d_deta_from_pfull(arr):
    """Compute $\partial/\partial\eta$ of the array on full hybrid levels.

    $\eta$ is the model vertical coordinate, and its value is assumed to simply
    increment by 1 from 0 at the surface upwards.  The data to be differenced
    is assumed to be defined at full pressure levels.

    Parameters
    ----------
    arr : xarray.DataArray containing the 'pfull' dim

    Returns
    -------
    deriv : xarray.DataArray with the derivative along 'pfull' computed via
            2nd order centered differencing.
    """  # noqa: W605
    right = arr[{internal_names.PFULL_STR: slice(2, None, None)}].values
    left = arr[{internal_names.PFULL_STR: slice(0, -2, 1)}].values
    deriv = xr.DataArray(np.zeros(arr.shape), dims=arr.dims,
                         coords=arr.coords)
    deriv[{internal_names.PFULL_STR: slice(1, -1, 1)}] = (right - left) / 2.
    deriv[{internal_names.PFULL_STR: 0}] = (
        arr[{internal_names.PFULL_STR: 1}].values -
        arr[{internal_names.PFULL_STR: 0}].values)
    deriv[{internal_names.PFULL_STR: -1}] = (
        arr[{internal_names.PFULL_STR: -1}].values -
        arr[{internal_names.PFULL_STR: -2}].values)
    return deriv


def dp_from_ps(bk, pk, ps, pfull_coord):
    """Compute pressure level thickness from surface pressure"""
    return d_deta_from_phalf(phalf_from_ps(bk, pk, ps), pfull_coord)


def integrate(arr, ddim, dim=False, is_pressure=False):
    """Integrate along the given dimension."""
    if is_pressure:
        dim = vert_coord_name(ddim)
    return (arr*ddim).sum(dim=dim)


def get_dim_name(arr, names):
    """Determine if an object has an attribute name matching a given list."""
    for name in names:
        # TODO: raise warning/exception when multiple names arr attrs.
        if hasattr(arr, name):
            return name
    raise AttributeError("No attributes of the object `{0}` match the "
                         "specified names of `{1}`".format(arr, names))


def vert_coord_name(arr):
    return get_dim_name(arr, [internal_names.PLEVEL_STR,
                              internal_names.PFULL_STR])


def int_dp_g(arr, dp):
    """Mass weighted integral."""
    return integrate(arr, to_pascal(dp, is_dp=True),
                     vert_coord_name(dp)) / GRAV_EARTH


def dp_from_p(p, ps, p_top=0., p_bot=1.1e5):
    """Get level thickness of pressure data, incorporating surface pressure.

    Level edges are defined as halfway between the levels, as well as the user-
    specified uppermost and lowermost values.  The dp of levels whose bottom
    pressure is less than the surface pressure is not changed by ps, since they
    don't intersect the surface.  If ps is in between a level's top and bottom
    pressures, then its dp becomes the pressure difference between its top and
    ps.  If ps is less than a level's top and bottom pressures, then that level
    is underground and its values are masked.

    Note that postprocessing routines (e.g. at GFDL) typically mask out data
    wherever the surface pressure is less than the level's given value, not the
    level's upper edge.  This masks out more levels than the

    """
    p_str = get_dim_name(p, (internal_names.PLEVEL_STR, 'plev'))
    p_vals = to_pascal(p.values.copy())

    # Layer edges are halfway between the given pressure levels.
    p_edges_interior = 0.5*(p_vals[:-1] + p_vals[1:])
    p_edges = np.concatenate(([p_bot], p_edges_interior, [p_top]))
    p_edge_above = p_edges[1:]
    p_edge_below = p_edges[:-1]
    dp = p_edge_below - p_edge_above
    if not all(np.sign(dp)):
        raise ValueError("dp array not all > 0 : {}".format(dp))
    # Pressure difference between ps and the upper edge of each pressure level.
    p_edge_above_xr = xr.DataArray(p_edge_above, dims=p.dims, coords=p.coords)
    dp_to_sfc = ps - p_edge_above_xr
    # Find the level adjacent to the masked, under-ground levels.
    change = xr.DataArray(np.zeros(dp_to_sfc.shape), dims=dp_to_sfc.dims,
                          coords=dp_to_sfc.coords)
    change[{p_str: slice(1, None)}] = np.diff(
        np.sign(ps - to_pascal(p.copy()))
    )
    dp_combined = xr.DataArray(np.where(change, dp_to_sfc, dp),
                               dims=dp_to_sfc.dims, coords=dp_to_sfc.coords)
    # Mask levels that are under ground.
    above_ground = ps > to_pascal(p.copy())
    above_ground[p_str] = p[p_str]
    dp_with_ps = dp_combined.where(above_ground)
    # Revert to original dim order.
    possible_dim_orders = [
        (internal_names.TIME_STR, p_str, internal_names.LAT_STR,
         internal_names.LON_STR),
        (internal_names.TIME_STR, p_str, internal_names.LAT_STR),
        (internal_names.TIME_STR, p_str, internal_names.LON_STR),
        (internal_names.TIME_STR, p_str),
        (p_str, internal_names.LAT_STR, internal_names.LON_STR),
        (p_str, internal_names.LAT_STR),
        (p_str, internal_names.LON_STR),
        (p_str,),
    ]
    for dim_order in possible_dim_orders:
        try:
            return dp_with_ps.transpose(*dim_order)
        except ValueError:
            logging.debug("Failed transpose to dims: {}".format(dim_order))
    else:
        logging.debug("No transpose was successful.")
        return dp_with_ps


def level_thickness(p, p_top=0., p_bot=1.01325e5):
    """
    Calculates the thickness, in Pa, of each pressure level.

    Assumes that the pressure values given are at the center of that model
    level, except for the lowest value (typically 1000 hPa), which is the
    bottom boundary. The uppermost level extends to 0 hPa.

    Unlike `dp_from_p`, this does not incorporate the surface pressure.

    """
    p_vals = to_pascal(p.values.copy())
    dp_vals = np.empty_like(p_vals)
    # Bottom level extends from p[0] to halfway betwen p[0] and p[1].
    dp_vals[0] = p_bot - 0.5*(p_vals[0] + p_vals[1])
    # Middle levels extend from halfway between [k-1], [k] and [k], [k+1].
    dp_vals[1:-1] = 0.5*(p_vals[0:-2] - p_vals[2:])
    # Top level extends from halfway between top two levels to 0 hPa.
    dp_vals[-1] = 0.5*(p_vals[-2] + p_vals[-1]) - p_top
    dp = p.copy()
    dp.values = dp_vals
    return dp


def does_coord_increase_w_index(arr):
    """Determine if the array values increase with the index.

    Useful, e.g., for pressure, which sometimes is indexed surface to TOA and
    sometimes the opposite.
    """
    diff = np.diff(arr)
    if not np.all(np.abs(np.sign(diff))):
        raise ValueError("Array is not monotonic: {}".format(arr))
    # Since we know its monotonic, just test the first value.
    return bool(diff[0])


bk = Var(
    name=internal_names.BK_STR,
    alt_names=internal_names.GRID_ATTRS[internal_names.BK_STR],
    def_vert=True,
    def_time=False,
    def_lon=False,
    def_lat=False
)


pk = Var(
    name=internal_names.PK_STR,
    alt_names=internal_names.GRID_ATTRS[internal_names.PK_STR],
    def_vert=True,
    def_time=False,
    def_lon=False,
    def_lat=False
)


pfull_coord = Var(
    name=internal_names.PFULL_STR,
    alt_names=internal_names.GRID_ATTRS[internal_names.PFULL_STR],
    def_vert=True,
    def_time=False,
    def_lon=False,
    def_lat=False
)


ps = Var(
    name='ps',
    domain='atmos',
    description='Surface pressure',
    units='Pa',
    def_vert=False,
    def_time=True,
    def_lon=True,
    def_lat=True
)


p_level = Var(
    name='p',
    alt_names=internal_names.GRID_ATTRS[internal_names.PLEVEL_STR],
    domain='atmos',
    description='Pressure on interpolated levels',
    units='Pa',
    def_vert=True,
    def_time=False,
    def_lon=False,
    def_lat=False
)


dp_level = Var(
    name='dp',
    description='Pressure thickness of model levels',
    units='Pa',
    def_vert=True,
    def_time=True,
    def_lon=True,
    def_lat=True,
    func=dp_from_p,
    variables=(p_level, ps)
)


p_eta = Var(
    name='p',
    description='Pressure at model-native level midpoints',
    units='Pa',
    def_vert=True,
    def_time=True,
    def_lon=True,
    def_lat=True,
    func=pfull_from_ps,
    variables=(bk, pk, ps, pfull_coord)
)


dp_eta = Var(
    name='dp',
    description='Pressure thickness of model levels',
    units='Pa',
    def_vert=True,
    def_time=True,
    def_lon=True,
    def_lat=True,
    func=dp_from_ps,
    variables=(bk, pk, ps, pfull_coord)
)



import os

from cftime import DatetimeNoLeap

from aospy import Proj, Model, Run, Var, Region
from aospy.data_loader import NestedDictDataLoader
from aospy.internal_names import LAND_MASK_STR, LON_STR


ROOT_PATH = os.path.dirname(__file__)


def total_precipitation(convection_rain, condensation_rain):
    return convection_rain + condensation_rain


precip_files = os.path.join(os.path.split(ROOT_PATH)[0], 'netcdf',
                            '000[4-6]0101.precip_monthly.nc')
sphum_files = os.path.join(os.path.split(ROOT_PATH)[0], 'netcdf',
                           '00060101.sphum_monthly.nc')
file_map = {'monthly': {'condensation_rain': precip_files,
                        'convection_rain': precip_files,
                        'sphum': sphum_files,
                        'ps': sphum_files}}
example_run = Run(
    name='example_run',
    description=(
        'Control simulation of the idealized moist model'
    ),
    data_loader=NestedDictDataLoader(file_map),
    default_start_date=DatetimeNoLeap(4, 1, 1),
    default_end_date=DatetimeNoLeap(6, 12, 31)
)

example_model = Model(
    name='example_model',
    grid_file_paths=(
        (os.path.join(os.path.split(ROOT_PATH)[0],
                      'netcdf', '00060101.sphum_monthly.nc'),
         os.path.join(os.path.split(ROOT_PATH)[0], 'netcdf',
                      'im.landmask.nc')),
    ),
    runs=[example_run],
    load_grid_data=True,
    grid_attrs={LAND_MASK_STR: 'custom_land_mask', LON_STR: 'custom_lon'}
)

example_proj = Proj(
    'example_proj',
    direc_out=os.path.join(os.path.dirname(__file__), 'test-files'),
    tar_direc_out=os.path.join(os.path.dirname(__file__), 'test-tar-files'),
    models=[example_model]
)

var_not_time_defined = Var(
    name='var_no_time_def',
    def_time=False,
)

condensation_rain = Var(
    name='condensation_rain',
    alt_names=('prec_ls',),
    def_time=True,
    description=('condensation rain'),
)

convection_rain = Var(
    name='convection_rain',
    alt_names=('prec_conv',),
    def_time=True,
    description=('convection rain'),
)

precip = Var(
    name='total_precipitation',
    def_time=True,
    description=('total precipitation rate'),
    func=total_precipitation,
    variables=(convection_rain, condensation_rain)
)

ps = Var(
    name='ps',
    def_time=True,
    description=('surface pressure'),
    def_vert=False
)

sphum = Var(
    name='sphum',
    def_time=True,
    description=('specific humidity'),
    def_vert=True
)

globe = Region(
    name='globe',
    description='Entire globe',
    west_bound=0,
    east_bound=360,
    south_bound=-90,
    north_bound=90,
    do_land_mask=False
)

sahel = Region(
    name='sahel',
    description='African Sahel',
    mask_bounds=[(0, 40, 10, 20),
                 (342, 360, 10, 20)],
    do_land_mask=True
)

bk = Var(
    name='bk',
    description=('hybrid coefficients'),
    def_time=False,
    def_vert=True,
    def_lon=False,
    def_lat=False
)

p = Var(
    name='p',
    description=('pressure'),
    def_time=True,
    def_vert=True,
    def_lon=True,
    def_lat=True
)

dp = Var(
    name='dp',
    description=('pressure thicknesses'),
    def_time=True,
    def_vert=True,
    def_lon=True,
    def_lat=True
)



"""Functionality for representing data on disk of individual models."""
import logging

import numpy as np
import xarray as xr

from ._constants import RADIUS_EARTH
from . import internal_names
from . import utils


def _get_grid_attr(grid_objs, attr_name):
    """Get attribute from the grid_objs file(s)."""
    for xds in grid_objs:
        try:
            return getattr(xds, attr_name)
        except AttributeError:
            pass


def _rename_coords(ds, attrs):
    """Rename coordinates to aospy's internal names."""
    for name_int, names_ext in attrs.items():
        # Check if coord is in dataset already.
        ds_coord_name = set(names_ext).intersection(set(ds.coords))
        if ds_coord_name:
            # Rename to the aospy internal name.
            try:
                ds = ds.rename({list(ds_coord_name)[0]: name_int})
                logging.debug("Rename coord from `{0}` to `{1}` for "
                              "Dataset `{2}`".format(ds_coord_name,
                                                     name_int, ds))
            # xarray throws a ValueError if the name already exists
            except ValueError:
                ds = ds
    return ds


def _bounds_from_array(arr, dim_name, bounds_name):
    """Get the bounds of an array given its center values.

    E.g. if lat-lon grid center lat/lon values are known, but not the
    bounds of each grid box.  The algorithm assumes that the bounds
    are simply halfway between each pair of center values.
    """
    # TODO: don't assume needed dimension is in axis=0
    # TODO: refactor to get rid of repetitive code
    spacing = arr.diff(dim_name).values
    lower = xr.DataArray(np.empty_like(arr), dims=arr.dims,
                         coords=arr.coords)
    lower.values[:-1] = arr.values[:-1] - 0.5*spacing
    lower.values[-1] = arr.values[-1] - 0.5*spacing[-1]
    upper = xr.DataArray(np.empty_like(arr), dims=arr.dims,
                         coords=arr.coords)
    upper.values[:-1] = arr.values[:-1] + 0.5*spacing
    upper.values[-1] = arr.values[-1] + 0.5*spacing[-1]
    bounds = xr.concat([lower, upper], dim='bounds')
    return bounds.T


def _diff_bounds(bounds, coord):
    """Get grid spacing by subtracting upper and lower bounds."""
    try:
        return bounds[:, 1] - bounds[:, 0]
    except IndexError:
        diff = np.diff(bounds, axis=0)
        return xr.DataArray(diff, dims=coord.dims, coords=coord.coords)


def _grid_sfc_area(lon, lat, lon_bounds=None, lat_bounds=None):
    """Calculate surface area of each grid cell in a lon-lat grid."""
    # Compute the bounds if not given.
    if lon_bounds is None:
        lon_bounds = _bounds_from_array(
            lon, internal_names.LON_STR, internal_names.LON_BOUNDS_STR)
    if lat_bounds is None:
        lat_bounds = _bounds_from_array(
            lat, internal_names.LAT_STR, internal_names.LAT_BOUNDS_STR)
    # Compute the surface area.
    dlon = _diff_bounds(utils.vertcoord.to_radians(lon_bounds, is_delta=True),
                        lon)
    sinlat_bounds = np.sin(utils.vertcoord.to_radians(lat_bounds,
                                                      is_delta=True))
    dsinlat = np.abs(_diff_bounds(sinlat_bounds, lat))
    sfc_area = dlon*dsinlat*(RADIUS_EARTH**2)
    # Rename the coordinates such that they match the actual lat / lon.
    try:
        sfc_area = sfc_area.rename(
            {internal_names.LAT_BOUNDS_STR: internal_names.LAT_STR,
             internal_names.LON_BOUNDS_STR: internal_names.LON_STR})
    except ValueError:
        pass
    # Clean up: correct names and dimension order.
    sfc_area = sfc_area.rename(internal_names.SFC_AREA_STR)
    sfc_area[internal_names.LAT_STR] = lat
    sfc_area[internal_names.LON_STR] = lon
    return sfc_area.transpose()


class Model(object):
    """An object that describes a single climate or weather model.

    Each `Model` object is associated with a parent `Proj` object and also with
    one or more child `Run` objects.

    If aospy is being used to work with non climate- or weather-model data, the
    `Model` object can be used e.g. to represent a gridded observational
    product, with its child `Run` objects representing different released
    versions of that dataset.

    Attributes
    ----------
    name : str
        The model's name
    description : str
        A description of the model
    proj : {None, aospy.Proj}
        The model's parent aospy.Proj object
    runs : list
        A list of this model's child Run objects
    default_runs : list
        The default subset of child run objects on which to perform
        calculations via `aospy.Calc` with this model if not otherwise
        specified
    grid_file_paths : list
        The paths to netCDF files stored on disk from which the model's
        coordinate data can be taken.
    default_start_date, default_end_date : datetime.datetime
        The default start and end dates of any calculations using this Model
    """

    def __init__(self, name=None, description=None, proj=None,
                 grid_file_paths=None, default_start_date=None,
                 default_end_date=None, runs=None, default_runs=None,
                 load_grid_data=False, grid_attrs=None):
        """
        Parameters
        ----------
        name : str
            The model's name.  This must be unique from that of any other
            `Model` objects being used by the parent `Proj`.
        description : str, optional
            A description of the model.  This is not used internally by
            aospy; it is solely for the user's information.
        proj : {None, aospy.Proj}, optional
            The parent Proj object.  When the parent `Proj` object is
            instantiated with this Model included in its `models` attribute,
            this will be over-written with that `Proj` object.
        grid_file_paths : {None, sequence of strings}, optional
            The paths to netCDF files stored on disk from which the model's
            coordinate data can be taken.
        default_start_date : {None, `datetime.datetime`}, optional
            Default start date of calculations to be performed using
            this Model.
        default_end_date : {None, `datetime.datetime`}, optional
            Default end date of calculations to be performed using
            this Model.
        runs : {None, sequence of aospy.Run objects}, optional
            The child run objects of this Model
        default_runs : {None, sequence of aospy.Run objects}, optional
            The subset of this Model's runs over which to perform calculations
            by default.
        load_grid_data : bool, optional (default False)
            Whether or not to load the grid data specified by 'grid_file_paths'
            upon initilization
        grid_attrs : dict, optional (default None)
            Dictionary mapping aospy internal names of grid attributes to their
            corresponding names used in a particular model.
            E.g. ``{TIME_STR: 'T'}``.  While aospy checks for a number of
            alternative names for grid attributes used by various models,
            it is not possible to anticipate all possible names.  This option
            allows the user to explicitly tell aospy which variables correspond
            to which internal names (internal names not provided in this
            dictionary will be attempted to be found in the usual way).  For a
            list of built-in alternative names see
            :ref:`the table here <built-in-alternative-names>`.

        See Also
        --------
        aospy.DataLoader, aospy.Proj, aospy.Run

        Notes
        -----
        A side-effect of instantiating a Model object is that the `parent`
        attribute of all of the model's `Run` objects is set to that model.

        """
        if isinstance(name, str) and name:
            self.name = name
        else:
            raise ValueError("Non-empty string value of `name` is required")
        self.description = '' if description is None else description
        self.proj = proj

        grid_file_paths = [] if grid_file_paths is None else grid_file_paths
        self.grid_file_paths = grid_file_paths

        self.default_start_date = default_start_date
        self.default_end_date = default_end_date

        self.runs = runs
        [setattr(run, 'parent', self) for run in self.runs]

        if default_runs is None:
            self.default_runs = []
        else:
            self.default_runs = default_runs

        self.grid_attrs = grid_attrs

        self._grid_data_is_set = False
        if load_grid_data:
            self.set_grid_data()
            self._grid_data_is_set = True

    def __str__(self):
        return 'Model instance "' + self.name + '"'

    __repr__ = __str__

    def _get_grid_files(self):
        """Get the files holding grid data for an aospy object."""
        grid_file_paths = self.grid_file_paths
        datasets = []
        if isinstance(grid_file_paths, str):
            grid_file_paths = [grid_file_paths]
        for path in grid_file_paths:
            try:
                ds = xr.open_dataset(path, decode_times=False)
            except TypeError:
                ds = xr.open_mfdataset(path, decode_times=False).load()
            except (RuntimeError, OSError) as e:
                msg = str(e) + ': {}'.format(path)
                raise RuntimeError(msg)
            datasets.append(ds)
        return tuple(datasets)

    def _set_mult_grid_attr(self):
        """
        Set multiple attrs from grid file given their names in the grid file.
        """
        grid_objs = self._get_grid_files()

        if self.grid_attrs is None:
            self.grid_attrs = {}

        # Override GRID_ATTRS with entries in grid_attrs
        attrs = internal_names.GRID_ATTRS.copy()
        for k, v in self.grid_attrs.items():
            if k not in attrs:
                raise ValueError(
                    'Unrecognized internal name, {!r}, specified for a '
                    'custom grid attribute name.  See the full list of '
                    'valid internal names below:\n\n{}'.format(
                        k, list(internal_names.GRID_ATTRS.keys())))
            attrs[k] = (v, )

        for name_int, names_ext in attrs.items():
            for name in names_ext:
                grid_attr = _get_grid_attr(grid_objs, name)
                if grid_attr is not None:
                    TIME_STR = internal_names.TIME_STR
                    renamed_attr = _rename_coords(grid_attr, attrs)
                    if ((TIME_STR not in renamed_attr.dims) and
                       (TIME_STR in renamed_attr.coords)):
                        renamed_attr = renamed_attr.drop(TIME_STR)
                    setattr(self, name_int, renamed_attr)
                    break

    def set_grid_data(self):
        """Populate the attrs that hold grid data."""
        if self._grid_data_is_set:
            return
        self._set_mult_grid_attr()
        if not np.any(getattr(self, 'sfc_area', None)):
            try:
                sfc_area = _grid_sfc_area(self.lon, self.lat, self.lon_bounds,
                                          self.lat_bounds)
            except AttributeError:
                sfc_area = _grid_sfc_area(self.lon, self.lat)
            self.sfc_area = sfc_area
        try:
            self.levs_thick = utils.vertcoord.level_thickness(self.level)
        except AttributeError:
            self.level = None
            self.levs_thick = None
        self._grid_data_is_set = True



#!/usr/bin/env python
"""Functionality relating to parsing and comparing longitudes."""

import numpy as np
import xarray as xr


def lon_to_0360(lon):
    """Convert longitude(s) to be within [0, 360).

    The Eastern hemisphere corresponds to 0 <= lon + (n*360) < 180, and the
    Western Hemisphere corresponds to 180 <= lon + (n*360) < 360, where 'n' is
    any integer (positive, negative, or zero).

    Parameters
    ----------
    lon : scalar or sequence of scalars
        One or more longitude values to be converted to lie in the [0, 360)
        range

    Returns
    -------
        If ``lon`` is a scalar, then a scalar of the same type in the range [0,
        360).  If ``lon`` is array-like, then an array-like of the same type
        with each element a scalar in the range [0, 360).

    """
    quotient = lon // 360
    return lon - quotient*360


def _lon_in_west_hem(lon):
    if lon_to_0360(lon) >= 180:
        return True
    else:
        return False


def lon_to_pm180(lon):
    """Convert longitude(s) to be within [-180, 180).

    The Eastern hemisphere corresponds to 0 <= lon + (n*360) < 180, and the
    Western Hemisphere corresponds to 180 <= lon + (n*360) < 360, where 'n' is
    any integer (positive, negative, or zero).

    Parameters
    ----------
    lon : scalar or sequence of scalars
        One or more longitude values to be converted to lie in the [-180, 180)
        range

    Returns
    -------
        If ``lon`` is a scalar, then a scalar of the same type in the range
        [-180, 180).  If ``lon`` is array-like, then an array-like of the same
        type with each element a scalar in the range [-180, 180).

    """
    lon0360 = lon_to_0360(lon)
    if _lon_in_west_hem(lon0360):
        return lon0360 - 360
    else:
        return lon0360


def _maybe_cast_to_lon(obj, strict=False):
    if isinstance(obj, Longitude):
        return obj
    try:
        return Longitude(obj)
    except (ValueError, TypeError) as e:
        if strict:
            raise type(e)(str(e))
        else:
            return obj


def _other_to_lon(func):
    """Wrapper for casting Longitude operator arguments to Longitude"""
    def func_other_to_lon(obj, other):
        return func(obj, _maybe_cast_to_lon(other))
    return func_other_to_lon


class Longitude(object):
    """Geographic longitude.

    Enables unambiguous comparison of longitudes using the standard comparison
    operators, regardless of they were initially represented with a 0 to 360
    convention, -180 to 180 convention, or anything else, and even if the
    original convention differs between them.

    Specifically, the ``<`` operator assesses if the first object is to the
    west of the second object, with the standard convention that longitudes in
    the Western Hemisphere are always to the west of longitudes in the Eastern
    Hemisphere.  The ``>`` operator is defined analogously.  ``==``, ``>=``,
    and ``<=`` are also all defined.

    In addition to other Longitude objects, the operators can be used to
    compare a Longitude object to anything that can be casted to a Longitude
    object, or to any sequence (e.g. a list or xarray.DataArray) whose elements
    can be casted to Longitude objects.

    """
    def __init__(self, value):
        """
        Parameters
        ----------
        value : {scalar, str}
            Scalars get converted to longitudes using the convention that 0-180
            corresponds to the Eastern Hemisphere, 180-360 corresponds to the
            Western Hemisphere, 360-540 the Eastern Hemisphere, and so on,
            including for negative numbers.

            Strings must be castable to a float or be a positive number in the
            range 0-180 followed by a single letter 'e' or 'w' (case
            insensitive).  For example, ``Longitude('10w')`` would yield a
            ``Longitude`` object corresponding to 10 degrees west longitude.
        """
        try:
            val_as_float = float(value)
        except (ValueError, TypeError):
            if not isinstance(value, str):
                raise ValueError('value must be a scalar or a string')
            if value[-1].lower() not in ('w', 'e'):
                raise ValueError("string inputs must end in 'e' or 'w'")
            try:
                lon_value = float(value[:-1])
            except ValueError:
                raise ValueError('improperly formatted string')
            if (lon_value < 0) or (lon_value > 180):
                raise ValueError('Value given as strings with hemisphere '
                                 'identifier must have numerical values '
                                 'within 0 and +180.  Value given: '
                                 '{}'.format(lon_value))
            self._longitude = lon_value
            self._hemisphere = value[-1].upper()
        else:
            lon_pm180 = lon_to_pm180(val_as_float)
            if _lon_in_west_hem(val_as_float):
                self._longitude = abs(lon_pm180)
                self._hemisphere = 'W'
            else:
                self._longitude = lon_pm180
                self._hemisphere = 'E'

    @property
    def longitude(self):
        """(scalar) The unsigned numerical value of the longitude.

        Always in the range 0 to 180.  Must be combined with the ``hemisphere``
        attribute to specify the exact latitude.

        """
        return self._longitude

    @longitude.setter
    def longitude(self, value):
        raise ValueError("'longitude' property cannot be modified after "
                         "Longitude object has been created.")

    @property
    def hemisphere(self):
        """{'W', 'E'} The longitude's hemisphere, either western or eastern."""
        return self._hemisphere

    @hemisphere.setter
    def hemisphere(self, value):
        raise ValueError("'hemisphere' property cannot be modified after "
                         "Longitude object has been created.")

    def __repr__(self):
        return "Longitude('{0}{1}')".format(self.longitude, self.hemisphere)

    @_other_to_lon
    def __eq__(self, other):
        if isinstance(other, Longitude):
            return (self.hemisphere == other.hemisphere and
                    self.longitude == other.longitude)
        else:
            return xr.apply_ufunc(np.equal, other, self)

    @_other_to_lon
    def __lt__(self, other):
        if isinstance(other, Longitude):
            if self.hemisphere == 'W':
                if other.hemisphere == 'E':
                    return True
                else:
                    return self.longitude > other.longitude
            else:
                if other.hemisphere == 'W':
                    return False
                else:
                    return self.longitude < other.longitude
        else:
            return xr.apply_ufunc(np.greater, other, self)

    @_other_to_lon
    def __gt__(self, other):
        if isinstance(other, Longitude):
            if self.hemisphere == 'W':
                if other.hemisphere == 'E':
                    return False
                else:
                    return self.longitude < other.longitude
            else:
                if other.hemisphere == 'W':
                    return True
                else:
                    return self.longitude > other.longitude
        else:
            return xr.apply_ufunc(np.less, other, self)

    @_other_to_lon
    def __le__(self, other):
        if isinstance(other, Longitude):
            return self < other or self == other
        else:
            return xr.apply_ufunc(np.greater_equal, other, self)

    @_other_to_lon
    def __ge__(self, other):
        if isinstance(other, Longitude):
            return self > other or self == other
        else:
            return xr.apply_ufunc(np.less_equal, other, self)

    def to_0360(self):
        """Convert longitude to its numerical value within [0, 360)."""
        if self.hemisphere == 'W':
            return -1*self.longitude + 360
        else:
            return self.longitude

    def to_pm180(self):
        """Convert longitude to its numerical value within [-180, 180)."""
        if self.hemisphere == 'W':
            return -1*self.longitude
        else:
            return self.longitude

    @_other_to_lon
    def __add__(self, other):
        return Longitude(self.to_0360() + other.to_0360())

    @_other_to_lon
    def __sub__(self, other):
        return Longitude(self.to_0360() - other.to_0360())


if __name__ == '__main__':
    pass



"""Subpackage comprising various utility functions used elsewhere in aospy."""
from . import io
from . import longitude
from .longitude import Longitude
from . import times
from . import vertcoord


__all__ = ['Longitude', 'io', 'longitude', 'times', 'vertcoord']



"""Physical constants used in aospy"""
import xarray as xr


RADIUS_EARTH = xr.DataArray(
    6370997.,
    attrs={'units': 'm',
           'description': 'Mean radius of Earth'}
)
GRAV_EARTH = xr.DataArray(
    9.81,
    attrs={'units': r'm s$^{-2}$',
           'description': 'Acceleration due to gravity'}
)



#!/usr/bin/env python
"""Test suite for aospy.utils module."""
import sys
import unittest

import numpy as np

import aospy.utils.vertcoord as vertcoord


class AospyUtilsTestCase(unittest.TestCase):
    def setUp(self):
        self.p_in_hpa = np.array([1000, 925, 850, 775, 700, 600, 500, 400, 300,
                                  200, 150, 100, 70, 50, 30, 20, 10],
                                 dtype=np.float64)
        self.p_in_pa = self.p_in_hpa*1e2
        self.p_top = 0
        self.p_bot = 1.1e5
        self.p_edges = 0.5*(self.p_in_pa[1:] + 0.5*self.p_in_pa[:-1])
        self.phalf = np.concatenate(([self.p_bot], self.p_edges, [self.p_top]))

    def tearDown(self):
        pass


class TestUtils(AospyUtilsTestCase):
    def test_to_pascal_scalar_positive(self):
        self.assertEqual(vertcoord.to_pascal(1e2), 1e4)
        self.assertEqual(vertcoord.to_pascal(1e5), 1e5)

    def test_to_pascal_scalar_negative(self):
        self.assertEqual(vertcoord.to_pascal(-1e2), -1e4)
        self.assertEqual(vertcoord.to_pascal(-1e5), -1e5)

    def test_to_pascal_array(self):
        np.testing.assert_array_equal(vertcoord.to_pascal(self.p_in_hpa),
                                      self.p_in_pa)
        np.testing.assert_array_equal(vertcoord.to_pascal(self.p_in_pa),
                                      self.p_in_pa)


if __name__ == '__main__':
    sys.exit(unittest.main())



#!/usr/bin/env python
import numpy as np
import pytest
import xarray as xr

from aospy.utils.longitude import Longitude, _maybe_cast_to_lon


_good_init_vals_attrs_objs = {
    -10: [10, 'W', Longitude('10W')],
    190.2: [169.8, 'W', Longitude('169.8W')],
    25: [25, 'E', Longitude('25E')],
    365: [5, 'E', Longitude('5E')],
    '45.5e': [45.5, 'E', Longitude('45.5E')],
    '22.2w': [22.2, 'W', Longitude('22.2W')],
    '0': [0, 'E', Longitude('0E')],
    }


_bad_init_vals = ['10ee', '-20e', '190w', None, 'abc', {'a': 1}]


@pytest.mark.parametrize(('val', 'attrs_and_obj'),
                         zip(_good_init_vals_attrs_objs.keys(),
                             _good_init_vals_attrs_objs.values()))
def test_longitude_init_good(val, attrs_and_obj):
    obj = Longitude(val)
    expected_lon = attrs_and_obj[0]
    expected_hem = attrs_and_obj[1]
    expected_obj = attrs_and_obj[2]
    assert obj.longitude == expected_lon
    assert obj.hemisphere == expected_hem
    assert Longitude(val) == expected_obj


def test_longitude_properties():
    lon = Longitude(5)
    with pytest.raises(ValueError):
        lon.longitude = 10
        lon.hemisphere = 'W'


@pytest.mark.parametrize(
    ('obj', 'expected_val'),
    [(Longitude('10w'), "Longitude('10.0W')"),
     (Longitude(0), "Longitude('0.0E')"),
     (Longitude(180), "Longitude('180.0W')")])
def test_longitude_repr(obj, expected_val):
    assert obj.__repr__() == expected_val


@pytest.mark.parametrize('bad_val', _bad_init_vals)
def test_longitude_init_bad(bad_val):
    with pytest.raises(ValueError):
        Longitude(bad_val)


@pytest.mark.parametrize('val', _good_init_vals_attrs_objs.keys())
def test_maybe_cast_to_lon_good(val):
    assert isinstance(_maybe_cast_to_lon(val), Longitude)


@pytest.mark.parametrize('bad_val', _bad_init_vals)
def test_maybe_cast_to_lon_bad(bad_val):
    assert isinstance(_maybe_cast_to_lon(bad_val), type(bad_val))
    with pytest.raises((ValueError, TypeError)):
        _maybe_cast_to_lon(bad_val, strict=True)


@pytest.mark.parametrize(
    ('obj1', 'obj2'),
    [(Longitude('100W'), Longitude('100W')),
     (Longitude('90E'), Longitude('90E')),
     (Longitude(0), Longitude(0)),
     (Longitude('0E'), 0),
     (Longitude('0E'), 720),
     (Longitude('0E'), [0, 720]),
     (Longitude('0E'), xr.DataArray([0, 720]))])
def test_lon_eq(obj1, obj2):
    assert np.all(obj1 == obj2)
    assert np.all(obj2 == obj1)


@pytest.mark.parametrize(
    ('obj1', 'obj2'),
    [(Longitude('100W'), Longitude('90W')),
     (Longitude('90E'), Longitude('100E')),
     (Longitude('10W'), Longitude('0E')),
     (Longitude('0E'), 10),
     (Longitude('0E'), [5, 10]),
     (Longitude('0E'), xr.DataArray([5, 10]))])
def test_lon_lt(obj1, obj2):
    assert np.all(obj1 < obj2)
    assert np.all(obj2 > obj1)


@pytest.mark.parametrize(
    ('obj1', 'obj2'),
    [(Longitude('90W'), Longitude('100W')),
     (Longitude('100E'), Longitude('90E')),
     (Longitude('0E'), Longitude('10W')),
     (Longitude('0E'), -10),
     (Longitude('0E'), [-10, -5]),
     (Longitude('0E'), xr.DataArray([-10, -5]))])
def test_lon_gt(obj1, obj2):
    assert np.all(obj1 > obj2)
    assert np.all(obj2 < obj1)


@pytest.mark.parametrize(
    ('obj1', 'obj2'),
    [(Longitude('100W'), Longitude('100W')),
     (Longitude('90E'), Longitude('90E')),
     (Longitude(0), Longitude(0)),
     (Longitude('100W'), Longitude('90W')),
     (Longitude('90E'), Longitude('100E')),
     (Longitude('10W'), Longitude('0E')),
     (Longitude('0E'), 10),
     (Longitude('0E'), [10, 0]),
     (Longitude('0E'), xr.DataArray([10, 0]))])
def test_lon_leq(obj1, obj2):
    assert np.all(obj1 <= obj2)
    assert np.all(obj2 >= obj2)


@pytest.mark.parametrize(
    ('obj1', 'obj2'),
    [(Longitude('100W'), Longitude('100W')),
     (Longitude('90E'), Longitude('90E')),
     (Longitude(0), Longitude(0)),
     (Longitude('90W'), Longitude('100W')),
     (Longitude('100E'), Longitude('90E')),
     (Longitude('0E'), Longitude('10W')),
     (Longitude('0E'), -10),
     (Longitude('0E'), [0, -10]),
     (Longitude('0E'), xr.DataArray([0, -10]))])
def test_lon_geq(obj1, obj2):
    assert np.all(obj1 >= obj2)
    assert np.all(obj2 <= obj1)


@pytest.mark.parametrize(
    ('obj', 'expected_val'),
    [(Longitude('100W'), 260),
     (Longitude(0), 0),
     (Longitude('20E'), 20)])
def test_to_0360(obj, expected_val):
    assert obj.to_0360() == expected_val


@pytest.mark.parametrize(
    ('obj', 'expected_val'),
    [(Longitude('100W'), -100),
     (Longitude(0), 0),
     (Longitude('20E'), 20)])
def test_to_pm180(obj, expected_val):
    assert obj.to_pm180() == expected_val


@pytest.mark.parametrize(
    ('obj1', 'obj2', 'expected_val'),
    [(Longitude(1), Longitude(1), Longitude(2)),
     (Longitude(175), Longitude(10), Longitude('175W'))])
def test_lon_add(obj1, obj2, expected_val):
    assert obj1 + obj2 == expected_val


@pytest.mark.parametrize(
    ('obj1', 'obj2', 'expected_val'),
    [(Longitude(1), Longitude(1), Longitude(0)),
     (Longitude(185), Longitude(10), Longitude('175E')),
     (Longitude(370), Longitude(20), Longitude('10W'))])
def test_lon_sub(obj1, obj2, expected_val):
    assert obj1 - obj2 == expected_val


if __name__ == '__main__':
    pass



print("import: 'aospy'")
import aospy

print("import: 'aospy.examples'")
import aospy.examples

print("import: 'aospy.test'")
import aospy.test

print("import: 'aospy.test.data'")
import aospy.test.data

print("import: 'aospy.test.data.objects'")
import aospy.test.data.objects

print("import: 'aospy.utils'")
import aospy.utils




#!/usr/bin/env python
"""Test suite for aospy.data_loader module."""
import datetime
import os
import unittest
import warnings

import numpy as np
import pytest
import xarray as xr

from cftime import DatetimeNoLeap

from aospy import Var
from aospy.data_loader import (DataLoader, DictDataLoader, GFDLDataLoader,
                               NestedDictDataLoader, grid_attrs_to_aospy_names,
                               set_grid_attrs_as_coords, _sel_var,
                               _prep_time_data,
                               _preprocess_and_rename_grid_attrs,
                               _maybe_cast_to_float64)
from aospy.internal_names import (LAT_STR, LON_STR, TIME_STR, TIME_BOUNDS_STR,
                                  BOUNDS_STR, SFC_AREA_STR, ETA_STR, PHALF_STR,
                                  TIME_WEIGHTS_STR, GRID_ATTRS, ZSURF_STR)
from aospy.utils import io
from .data.objects.examples import (condensation_rain, convection_rain, precip,
                                    file_map, ROOT_PATH, example_model, bk)


def _open_ds_catch_warnings(path):
    with warnings.catch_warnings(record=True):
        return xr.open_dataset(path)


@pytest.mark.parametrize(
    ('input_dtype', 'expected_dtype'),
    [(np.float32, np.float64),
     (np.int, np.int)])
def test_maybe_cast_to_float64(input_dtype, expected_dtype):
    da = xr.DataArray(np.ones(3, dtype=input_dtype))
    result = _maybe_cast_to_float64(da).dtype
    assert result == expected_dtype


_DATE_RANGES = {
    'datetime': (datetime.datetime(2000, 1, 1),
                 datetime.datetime(2002, 12, 31)),
    'datetime64': (np.datetime64('2000-01-01'),
                   np.datetime64('2002-12-31')),
    'cftime': (DatetimeNoLeap(2000, 1, 1),
               DatetimeNoLeap(2002, 12, 31)),
    'str': ('2000', '2002')
}


@pytest.fixture(params=_DATE_RANGES.values(), ids=list(_DATE_RANGES.keys()))
def generate_file_set_args(request):
    start_date, end_date = request.param
    return dict(
        var=condensation_rain, start_date=start_date, end_date=end_date,
        domain='atmos', intvl_in='monthly', dtype_in_vert='sigma',
        dtype_in_time='ts', intvl_out=None)


@pytest.fixture()
def alt_lat_str():
    return 'LATITUDE'


@pytest.fixture()
def var_name():
    return 'a'


@pytest.fixture()
def ds(alt_lat_str, var_name):
    time_bounds = np.array([[0, 31], [31, 59], [59, 90]])
    bounds = np.array([0, 1])
    time = np.array([15, 46, 74])
    data = np.zeros((3, 1, 1))
    lat = [0]
    lon = [0]
    ds = xr.DataArray(data,
                      coords=[time, lat, lon],
                      dims=[TIME_STR, alt_lat_str, LON_STR],
                      name=var_name).to_dataset()
    ds[TIME_BOUNDS_STR] = xr.DataArray(time_bounds,
                                       coords=[time, bounds],
                                       dims=[TIME_STR, BOUNDS_STR],
                                       name=TIME_BOUNDS_STR)
    units_str = 'days since 2000-01-01 00:00:00'
    ds[TIME_STR].attrs['units'] = units_str
    ds[TIME_BOUNDS_STR].attrs['units'] = units_str
    return ds


@pytest.fixture()
def inst_ds(ds):
    inst_time = np.array([3, 6, 9])
    inst_units_str = 'hours since 2000-01-01 00:00:00'
    inst_ds = ds.copy()
    inst_ds.drop(TIME_BOUNDS_STR)
    inst_ds[TIME_STR].values = inst_time
    inst_ds[TIME_STR].attrs['units'] = inst_units_str
    return inst_ds


def _gfdl_data_loader_kwargs(data_start_date, data_end_date):
    return dict(data_direc=os.path.join('.', 'test'),
                data_dur=6,
                data_start_date=data_start_date,
                data_end_date=data_end_date,
                upcast_float32=False,
                data_vars='minimal',
                coords='minimal')


_DATA_LOADER_KWARGS = {
    'DataLoader': (DataLoader, {}),
    'DictDataLoader': (DictDataLoader, dict(file_map={'monthly': ['a.nc']})),
    'NestedDictDataLoader': (
        NestedDictDataLoader,
        dict(file_map={'monthly': {'condensation_rain': ['a.nc']}})),
    'GFDLDataLoader-datetime': (
        GFDLDataLoader, _gfdl_data_loader_kwargs(
            datetime.datetime(2000, 1, 1), datetime.datetime(2012, 12, 31))),
    'GFDLDataLoader-datetime64': (
        GFDLDataLoader, _gfdl_data_loader_kwargs(
            np.datetime64('2000-01-01'), np.datetime64('2012-12-31'))),
    'GFDLDataLoader-cftime': (
        GFDLDataLoader, _gfdl_data_loader_kwargs(
            DatetimeNoLeap(2000, 1, 1), DatetimeNoLeap(2012, 12, 31))),
    'GFDLDataLoader-str': (
        GFDLDataLoader, _gfdl_data_loader_kwargs('2000', '2012'))
}


@pytest.fixture(params=_DATA_LOADER_KWARGS.values(),
                ids=list(_DATA_LOADER_KWARGS.keys()))
def data_loader(request):
    data_loader_type, kwargs = request.param
    return data_loader_type(**kwargs)


_GFDL_DATA_LOADER_KWARGS = {key: _DATA_LOADER_KWARGS[key] for key in
                            _DATA_LOADER_KWARGS if 'GFDL' in key}


@pytest.fixture(params=_GFDL_DATA_LOADER_KWARGS.values(),
                ids=list(_GFDL_DATA_LOADER_KWARGS.keys()))
def gfdl_data_loader(request):
    data_loader_type, kwargs = request.param
    return data_loader_type(**kwargs)


def test_rename_grid_attrs_ds(ds, alt_lat_str):
    assert LAT_STR not in ds
    assert alt_lat_str in ds
    ds = grid_attrs_to_aospy_names(ds)
    assert LAT_STR in ds


def test_rename_grid_attrs_dim_no_coord(ds, var_name):
    bounds_dim = 'nv'
    assert bounds_dim not in ds
    assert bounds_dim in GRID_ATTRS[BOUNDS_STR]
    # Create DataArray with all dims lacking coords
    values = ds[var_name].values
    arr = xr.DataArray(values, name='dummy')
    # Insert name to be replaced (its physical meaning doesn't matter here)
    ds = arr.rename({'dim_0': bounds_dim}).to_dataset()
    assert not ds[bounds_dim].coords
    result = grid_attrs_to_aospy_names(ds)
    assert not result[BOUNDS_STR].coords


def test_rename_grid_attrs_skip_scalar_dim(ds):
    phalf_dim = 'phalf'
    assert phalf_dim not in ds
    assert phalf_dim in GRID_ATTRS[PHALF_STR]
    ds_copy = ds.copy()
    ds_copy[phalf_dim] = 4
    ds_copy = ds_copy.set_coords(phalf_dim)
    result = grid_attrs_to_aospy_names(ds_copy)
    xr.testing.assert_identical(result[phalf_dim], ds_copy[phalf_dim])


def test_rename_grid_attrs_copy_attrs(ds, alt_lat_str):
    orig_attrs = {'dummy_key': 'dummy_val'}
    ds_orig = ds.copy()
    ds_orig[alt_lat_str].attrs = orig_attrs
    ds = grid_attrs_to_aospy_names(ds_orig)
    assert ds[LAT_STR].attrs == orig_attrs


def test_rename_grid_attrs_custom(ds, alt_lat_str):
    assert LAT_STR not in ds
    ds = ds.rename({alt_lat_str: 'custom_lat_name'})
    ds = grid_attrs_to_aospy_names(ds, {LAT_STR: 'custom_lat_name'})
    assert LAT_STR in ds
    assert 'custom_lat_name' not in ds


def test_rename_grid_attrs_custom_error(ds, alt_lat_str):
    assert LAT_STR not in ds
    ds = ds.rename({alt_lat_str: 'custom_lat_name'})
    with pytest.raises(ValueError):
        ds = grid_attrs_to_aospy_names(ds, {alt_lat_str: 'custom_lat_name'})


def test_set_grid_attrs_as_coords(ds, var_name):
    ds = grid_attrs_to_aospy_names(ds)
    sfc_area = ds[var_name].isel(**{TIME_STR: 0}).drop(TIME_STR)
    ds[SFC_AREA_STR] = sfc_area

    assert SFC_AREA_STR not in ds.coords

    ds = set_grid_attrs_as_coords(ds)
    assert SFC_AREA_STR in ds.coords
    assert TIME_BOUNDS_STR in ds.coords


def test_sel_var():
    time = np.array([0, 31, 59]) + 15
    data = np.zeros((3))
    ds = xr.DataArray(data,
                      coords=[time],
                      dims=[TIME_STR],
                      name=convection_rain.name).to_dataset()
    condensation_rain_alt_name, = condensation_rain.alt_names
    ds[condensation_rain_alt_name] = xr.DataArray(data, coords=[ds.time])
    result = _sel_var(ds, convection_rain)
    assert result.name == convection_rain.name

    result = _sel_var(ds, condensation_rain)
    assert result.name == condensation_rain.name

    with pytest.raises(LookupError):
        _sel_var(ds, precip)


def test_maybe_apply_time_shift(data_loader, ds, inst_ds, var_name,
                                generate_file_set_args):
    ds = xr.decode_cf(ds)
    da = ds[var_name]

    result = data_loader._maybe_apply_time_shift(
        da.copy(), **generate_file_set_args)[TIME_STR]
    assert result.identical(da[TIME_STR])

    offset = data_loader._maybe_apply_time_shift(
        da.copy(), {'days': 1}, **generate_file_set_args)
    result = offset[TIME_STR]

    expected = da[TIME_STR] + np.timedelta64(1, 'D')
    expected[TIME_STR] = expected

    assert result.identical(expected)


def test_maybe_apply_time_shift_ts(gfdl_data_loader, ds, var_name,
                                   generate_file_set_args):
    ds = xr.decode_cf(ds)
    da = ds[var_name]
    result = gfdl_data_loader._maybe_apply_time_shift(
        da.copy(), **generate_file_set_args)[TIME_STR]
    assert result.identical(da[TIME_STR])


def test_maybe_apply_time_shift_inst(gfdl_data_loader, inst_ds, var_name,
                                     generate_file_set_args):
    inst_ds = xr.decode_cf(inst_ds)
    generate_file_set_args['dtype_in_time'] = 'inst'
    generate_file_set_args['intvl_in'] = '3hr'
    da = inst_ds[var_name]
    result = gfdl_data_loader._maybe_apply_time_shift(
        da.copy(), **generate_file_set_args)[TIME_STR]

    expected = da[TIME_STR] + np.timedelta64(-3, 'h')
    expected[TIME_STR] = expected
    assert result.identical(expected)

    generate_file_set_args['intvl_in'] = 'daily'
    da = inst_ds[var_name]
    result = gfdl_data_loader._maybe_apply_time_shift(
        da.copy(), **generate_file_set_args)[TIME_STR]

    expected = da[TIME_STR]
    expected[TIME_STR] = expected
    assert result.identical(expected)


def test_prep_time_data(inst_ds):
    assert (TIME_WEIGHTS_STR not in inst_ds)
    ds = _prep_time_data(inst_ds)
    assert (TIME_WEIGHTS_STR in ds)


def test_preprocess_and_rename_grid_attrs(ds, alt_lat_str):
    def preprocess_func(ds, **kwargs):
        # Corrupt a grid attribute name so that we test
        # that grid_attrs_to_aospy_names is still called
        # after
        ds = ds.rename({LON_STR: 'LONGITUDE'})
        ds.attrs['a'] = 'b'
        return ds

    assert LAT_STR not in ds
    assert alt_lat_str in ds
    assert LON_STR in ds

    expected = ds.rename({alt_lat_str: LAT_STR})
    expected = expected.set_coords(TIME_BOUNDS_STR)
    expected.attrs['a'] = 'b'
    result = _preprocess_and_rename_grid_attrs(preprocess_func)(ds)
    xr.testing.assert_identical(result, expected)


def test_generate_file_set(data_loader, generate_file_set_args):
    if type(data_loader) is DataLoader:
        with pytest.raises(NotImplementedError):
            data_loader._generate_file_set()

    elif isinstance(data_loader, DictDataLoader):
        result = data_loader._generate_file_set(
            **generate_file_set_args)
        expected = ['a.nc']
        result == expected

        with pytest.raises(KeyError):
            generate_file_set_args['intvl_in'] = 'daily'
            result = data_loader._generate_file_set(
                **generate_file_set_args)

    elif isinstance(data_loader, NestedDictDataLoader):
        result = data_loader._generate_file_set(
            **generate_file_set_args)
        expected = ['a.nc']
        assert result == expected

        with pytest.raises(KeyError):
            generate_file_set_args['var'] = convection_rain
            result = data_loader._generate_file_set(
                **generate_file_set_args)

    else:
        with pytest.raises(IOError):
            data_loader._generate_file_set(**generate_file_set_args)


def test_overriding_constructor(gfdl_data_loader, ds):
    new = GFDLDataLoader(gfdl_data_loader,
                         data_direc=os.path.join('.', 'a'))
    assert new.data_direc == os.path.join('.', 'a')
    assert new.data_dur == gfdl_data_loader.data_dur
    assert new.data_start_date == gfdl_data_loader.data_start_date
    assert new.data_end_date == gfdl_data_loader.data_end_date
    assert new.preprocess_func == gfdl_data_loader.preprocess_func
    assert new.upcast_float32 == gfdl_data_loader.upcast_float32

    new = GFDLDataLoader(gfdl_data_loader, data_dur=8)
    assert new.data_dur == 8

    new = GFDLDataLoader(gfdl_data_loader,
                         data_start_date=datetime.datetime(2001, 1, 1))
    assert new.data_start_date == datetime.datetime(2001, 1, 1)

    new = GFDLDataLoader(gfdl_data_loader,
                         data_end_date=datetime.datetime(2003, 12, 31))
    assert new.data_end_date == datetime.datetime(2003, 12, 31)

    new = GFDLDataLoader(gfdl_data_loader, preprocess_func=lambda ds: ds)
    xr.testing.assert_identical(new.preprocess_func(ds), ds)

    new = GFDLDataLoader(gfdl_data_loader, upcast_float32=True)
    assert new.upcast_float32

    new = GFDLDataLoader(gfdl_data_loader, data_vars='all')
    assert new.data_vars == 'all'

    new = GFDLDataLoader(gfdl_data_loader, coords='all')
    assert new.coords == 'all'


_GFDL_DATE_RANGES = {
    'datetime': (datetime.datetime(2010, 1, 1),
                 datetime.datetime(2010, 12, 31)),
    'datetime64': (np.datetime64('2010-01-01'),
                   np.datetime64('2010-12-31')),
    'cftime': (DatetimeNoLeap(2010, 1, 1),
               DatetimeNoLeap(2010, 12, 31)),
    'str': ('2010', '2010')
}


@pytest.mark.parametrize(['start_date', 'end_date'],
                         _GFDL_DATE_RANGES.values(),
                         ids=list(_GFDL_DATE_RANGES.keys()))
def test_input_data_paths_gfdl(gfdl_data_loader, start_date, end_date):
    expected = [os.path.join('.', 'test', 'atmos', 'ts', 'monthly', '6yr',
                             'atmos.200601-201112.temp.nc')]
    result = gfdl_data_loader._input_data_paths_gfdl(
        'temp', start_date, end_date, 'atmos',
        'monthly', 'pressure', 'ts', None)
    assert result == expected

    expected = [os.path.join('.', 'test', 'atmos_daily', 'ts', 'daily',
                             '6yr',
                             'atmos_daily.20060101-20111231.temp.nc')]
    result = gfdl_data_loader._input_data_paths_gfdl(
        'temp', start_date, end_date, 'atmos',
        'daily', 'pressure', 'ts', None)
    assert result == expected

    expected = [os.path.join('.', 'test', 'atmos_daily', 'ts', 'daily',
                             '6yr',
                             'atmos_daily.20060101-20111231.temp.nc')]
    result = gfdl_data_loader._input_data_paths_gfdl(
        'temp', start_date, end_date, 'atmos',
        'daily', 'pressure', 'ts', None)
    assert result == expected

    expected = [os.path.join('.', 'test', 'atmos_level', 'ts', 'monthly',
                             '6yr', 'atmos_level.200601-201112.temp.nc')]
    result = gfdl_data_loader._input_data_paths_gfdl(
        'temp', start_date, end_date, 'atmos',
        'monthly', ETA_STR, 'ts', None)
    assert result == expected

    expected = [os.path.join('.', 'test', 'atmos', 'ts', 'monthly',
                             '6yr', 'atmos.200601-201112.ps.nc')]
    result = gfdl_data_loader._input_data_paths_gfdl(
        'ps', start_date, end_date, 'atmos',
        'monthly', ETA_STR, 'ts', None)
    assert result == expected

    expected = [os.path.join('.', 'test', 'atmos_inst', 'ts', 'monthly',
                             '6yr', 'atmos_inst.200601-201112.temp.nc')]
    result = gfdl_data_loader._input_data_paths_gfdl(
        'temp', start_date, end_date, 'atmos',
        'monthly', 'pressure', 'inst', None)
    assert result == expected

    expected = [os.path.join('.', 'test', 'atmos', 'av', 'monthly_6yr',
                             'atmos.2006-2011.jja.nc')]
    result = gfdl_data_loader._input_data_paths_gfdl(
        'temp', start_date, end_date, 'atmos',
        'monthly', 'pressure', 'av', 'jja')
    assert result == expected


# TODO: Parametrize these tests
def test_data_name_gfdl_annual():
    for data_type in ['ts', 'inst']:
        expected = 'atmos.2010.temp.nc'
        result = io.data_name_gfdl('temp', 'atmos', data_type,
                                   'annual', 2010, None, 2000, 1)
        assert result == expected

        expected = 'atmos.2006-2011.temp.nc'
        result = io.data_name_gfdl('temp', 'atmos', data_type,
                                   'annual', 2010, None, 2000, 6)
        assert result == expected

    for intvl_type in ['annual', 'ann']:
        expected = 'atmos.2010.ann.nc'
        result = io.data_name_gfdl('temp', 'atmos', 'av',
                                   intvl_type, 2010, None, 2000, 1)
        assert result == expected

        expected = 'atmos.2006-2011.ann.nc'
        result = io.data_name_gfdl('temp', 'atmos', 'av',
                                   intvl_type, 2010, None, 2000, 6)
        assert result == expected

    expected = 'atmos.2006-2011.01-12.nc'
    result = io.data_name_gfdl('temp', 'atmos', 'av_ts',
                               'annual', 2010, None, 2000, 6)
    assert result == expected


def test_data_name_gfdl_monthly():
    for data_type in ['ts', 'inst']:
        expected = 'atmos.200601-201112.temp.nc'
        result = io.data_name_gfdl('temp', 'atmos', data_type,
                                   'monthly', 2010, 'jja', 2000, 6)
        assert result == expected

    for intvl_type in ['monthly', 'mon']:
        expected = 'atmos.2010.jja.nc'
        result = io.data_name_gfdl('temp', 'atmos', 'av',
                                   intvl_type, 2010, 'jja', 2000, 1)
        assert result == expected

        expected = 'atmos.2006-2011.jja.nc'
        result = io.data_name_gfdl('temp', 'atmos', 'av',
                                   intvl_type, 2010, 'jja', 2000, 6)
        assert result == expected

    expected = 'atmos.2006-2011.01-12.nc'
    result = io.data_name_gfdl('temp', 'atmos', 'av_ts',
                               'monthly', 2010, 'jja', 2000, 6)
    assert result == expected


def test_data_name_gfdl_daily():
    for data_type in ['ts', 'inst']:
        expected = 'atmos.20060101-20111231.temp.nc'
        result = io.data_name_gfdl('temp', 'atmos', data_type,
                                   'daily', 2010, None, 2000, 6)
        assert result == expected

    with pytest.raises(NameError):
        io.data_name_gfdl('temp', 'atmos', 'av',
                          'daily', 2010, None, 2000, 6)

    expected = 'atmos.2006-2011.01-12.nc'
    result = io.data_name_gfdl('temp', 'atmos', 'av_ts',
                               'daily', 2010, None, 2000, 6)
    assert result == expected


def test_data_name_gfdl_hr():
    for data_type in ['ts', 'inst']:
        expected = 'atmos.2006010100-2011123123.temp.nc'
        result = io.data_name_gfdl('temp', 'atmos', data_type,
                                   '3hr', 2010, None, 2000, 6)
        assert result == expected

    with pytest.raises(NameError):
        io.data_name_gfdl('temp', 'atmos', 'av',
                          '3hr', 2010, None, 2000, 6)

    expected = 'atmos.2006-2011.01-12.nc'
    result = io.data_name_gfdl('temp', 'atmos', 'av_ts',
                               '3hr', 2010, None, 2000, 6)
    assert result == expected


def test_data_name_gfdl_seasonal():
    for data_type in ['ts', 'inst']:
        with pytest.raises(NameError):
            io.data_name_gfdl('temp', 'atmos', data_type,
                              'seasonal', 2010, None, 2000, 6)

    for intvl_type in ['seasonal', 'seas']:
        expected = 'atmos.2010.JJA.nc'
        result = io.data_name_gfdl('temp', 'atmos', 'av',
                                   intvl_type, 2010, 'jja', 2000, 1)
        assert result == expected

        expected = 'atmos.2006-2011.JJA.nc'
        result = io.data_name_gfdl('temp', 'atmos', 'av',
                                   intvl_type, 2010, 'jja', 2000, 6)
        assert result == expected

    expected = 'atmos.2006-2011.01-12.nc'
    result = io.data_name_gfdl('temp', 'atmos', 'av_ts',
                               'seasonal', 2010, None, 2000, 6)
    assert result == expected


@pytest.fixture()
def load_variable_data_loader():
    return NestedDictDataLoader(file_map, upcast_float32=False)


_LOAD_VAR_DATE_RANGES = {
    'datetime': (datetime.datetime(5, 1, 1),
                 datetime.datetime(5, 12, 31)),
    'datetime64': (np.datetime64('0005-01-01'),
                   np.datetime64('0005-12-31')),
    'cftime': (DatetimeNoLeap(5, 1, 1), DatetimeNoLeap(5, 12, 31)),
    'str': ('0005', '0005')
}


@pytest.mark.parametrize(['start_date', 'end_date'],
                         _LOAD_VAR_DATE_RANGES.values(),
                         ids=list(_LOAD_VAR_DATE_RANGES.keys()))
def test_load_variable(load_variable_data_loader, start_date, end_date):
    result = load_variable_data_loader.load_variable(
        condensation_rain, start_date, end_date, intvl_in='monthly')
    filepath = os.path.join(os.path.split(ROOT_PATH)[0], 'netcdf',
                            '00050101.precip_monthly.nc')
    expected = _open_ds_catch_warnings(filepath)['condensation_rain']
    np.testing.assert_array_equal(result.values, expected.values)


@pytest.mark.filterwarnings('ignore:CFTimeIndex.data is deprecated')
@pytest.mark.parametrize(['start_date', 'end_date'],
                         _LOAD_VAR_DATE_RANGES.values(),
                         ids=list(_LOAD_VAR_DATE_RANGES.keys()))
def test_load_variable_does_not_warn(load_variable_data_loader,
                                     start_date, end_date):
    with warnings.catch_warnings(record=True) as warnlog:
        load_variable_data_loader.load_variable(
            condensation_rain,
            start_date, end_date,
            intvl_in='monthly')
    assert len(warnlog) == 0


@pytest.mark.parametrize(['start_date', 'end_date'],
                         _LOAD_VAR_DATE_RANGES.values(),
                         ids=list(_LOAD_VAR_DATE_RANGES.keys()))
def test_load_variable_float32_to_float64(load_variable_data_loader,
                                          start_date, end_date):
    def preprocess(ds, **kwargs):
        # This function converts testing data to the float32 datatype
        return ds.astype(np.float32)
    load_variable_data_loader.upcast_float32 = True
    load_variable_data_loader.preprocess_func = preprocess
    result = load_variable_data_loader.load_variable(
        condensation_rain, start_date,
        end_date,
        intvl_in='monthly').dtype
    expected = np.float64
    assert result == expected


@pytest.mark.parametrize(['start_date', 'end_date'],
                         _LOAD_VAR_DATE_RANGES.values(),
                         ids=list(_LOAD_VAR_DATE_RANGES.keys()))
def test_load_variable_maintain_float32(load_variable_data_loader,
                                        start_date, end_date):
    def preprocess(ds, **kwargs):
        # This function converts testing data to the float32 datatype
        return ds.astype(np.float32)
    load_variable_data_loader.preprocess_func = preprocess
    load_variable_data_loader.upcast_float32 = False
    result = load_variable_data_loader.load_variable(
        condensation_rain, start_date,
        end_date,
        intvl_in='monthly').dtype
    expected = np.float32
    assert result == expected


_LOAD_VAR_MULTI_YEAR_RANGES = {
    'datetime': (datetime.datetime(4, 1, 1),
                 datetime.datetime(5, 12, 31)),
    'datetime64': (np.datetime64('0004-01-01'),
                   np.datetime64('0005-12-31')),
    'cftime': (DatetimeNoLeap(4, 1, 1), DatetimeNoLeap(5, 12, 31)),
    'str': ('0004', '0005')
}


@pytest.mark.parametrize(['start_date', 'end_date'],
                         _LOAD_VAR_MULTI_YEAR_RANGES.values(),
                         ids=list(_LOAD_VAR_MULTI_YEAR_RANGES.keys()))
def test_load_variable_data_vars_all(load_variable_data_loader,
                                     start_date, end_date):
    def preprocess(ds, **kwargs):
        # This function drops the time coordinate from condensation_rain
        temp = ds[condensation_rain.name]
        temp = temp.isel(time=0, drop=True)
        ds = ds.drop(condensation_rain.name)
        ds[condensation_rain.name] = temp
        assert TIME_STR not in ds[condensation_rain.name].coords
        return ds

    load_variable_data_loader.data_vars = 'all'
    load_variable_data_loader.preprocess_func = preprocess
    data = load_variable_data_loader.load_variable(
        condensation_rain, start_date, end_date,
        intvl_in='monthly')
    assert TIME_STR in data.coords


@pytest.mark.parametrize(['start_date', 'end_date'],
                         _LOAD_VAR_MULTI_YEAR_RANGES.values(),
                         ids=list(_LOAD_VAR_MULTI_YEAR_RANGES.keys()))
def test_load_variable_data_vars_default(load_variable_data_loader, start_date,
                                         end_date):
    data = load_variable_data_loader.load_variable(
        condensation_rain, start_date, end_date,
        intvl_in='monthly')
    assert TIME_STR in data.coords


@pytest.mark.parametrize(['start_date', 'end_date'],
                         _LOAD_VAR_MULTI_YEAR_RANGES.values(),
                         ids=list(_LOAD_VAR_MULTI_YEAR_RANGES.keys()))
def test_load_variable_coords_all(load_variable_data_loader, start_date,
                                  end_date):
    load_variable_data_loader.coords = 'all'
    data = load_variable_data_loader.load_variable(
        condensation_rain, start_date, end_date,
        intvl_in='monthly')
    assert TIME_STR in data[ZSURF_STR].coords


@pytest.mark.parametrize('year', [4, 5, 6])
def test_load_variable_non_0001_refdate(load_variable_data_loader, year):
    def preprocess(ds, **kwargs):
        # This function converts our testing data (encoded with a units
        # attribute with a reference data of 0001-01-01) to one
        # with a reference data of 0004-01-01 (to do so we also need
        # to offset the raw time values by three years).
        three_yrs = 1095.
        ds['time'] = ds['time'] - three_yrs
        ds['time'].attrs['units'] = 'days since 0004-01-01 00:00:00'
        ds['time'].attrs['calendar'] = 'noleap'
        ds['time_bounds'] = ds['time_bounds'] - three_yrs
        ds['time_bounds'].attrs['units'] = 'days since 0004-01-01 00:00:00'
        ds['time_bounds'].attrs['calendar'] = 'noleap'
        return ds

    load_variable_data_loader.preprocess_func = preprocess

    result = load_variable_data_loader.load_variable(
        condensation_rain, DatetimeNoLeap(year, 1, 1),
        DatetimeNoLeap(year, 12, 31),
        intvl_in='monthly')
    filepath = os.path.join(os.path.split(ROOT_PATH)[0], 'netcdf',
                            '000{}0101.precip_monthly.nc'.format(year))
    expected = _open_ds_catch_warnings(filepath)['condensation_rain']
    np.testing.assert_allclose(result.values, expected.values)


def test_load_variable_preprocess(load_variable_data_loader):
    def preprocess(ds, **kwargs):
        if kwargs['start_date'] == DatetimeNoLeap(5, 1, 1):
            ds['condensation_rain'] = 10. * ds['condensation_rain']
        return ds

    load_variable_data_loader.preprocess_func = preprocess

    result = load_variable_data_loader.load_variable(
        condensation_rain, DatetimeNoLeap(5, 1, 1),
        DatetimeNoLeap(5, 12, 31),
        intvl_in='monthly')
    filepath = os.path.join(os.path.split(ROOT_PATH)[0], 'netcdf',
                            '00050101.precip_monthly.nc')
    expected = 10. * _open_ds_catch_warnings(filepath)['condensation_rain']
    np.testing.assert_allclose(result.values, expected.values)

    result = load_variable_data_loader.load_variable(
        condensation_rain, DatetimeNoLeap(4, 1, 1),
        DatetimeNoLeap(4, 12, 31),
        intvl_in='monthly')
    filepath = os.path.join(os.path.split(ROOT_PATH)[0], 'netcdf',
                            '00040101.precip_monthly.nc')
    expected = _open_ds_catch_warnings(filepath)['condensation_rain']
    np.testing.assert_allclose(result.values, expected.values)


def test_load_variable_mask_and_scale(load_variable_data_loader):
    def convert_all_to_missing_val(ds, **kwargs):
        ds['condensation_rain'] = 0. * ds['condensation_rain'] + 1.0e20
        ds['condensation_rain'].attrs['_FillValue'] = 1.0e20
        return ds

    load_variable_data_loader.preprocess_func = convert_all_to_missing_val

    data = load_variable_data_loader.load_variable(
        condensation_rain, DatetimeNoLeap(5, 1, 1),
        DatetimeNoLeap(5, 12, 31),
        intvl_in='monthly')

    num_non_missing = np.isfinite(data).sum().item()
    expected_num_non_missing = 0
    assert num_non_missing == expected_num_non_missing


def test_recursively_compute_variable_native(load_variable_data_loader):
    result = load_variable_data_loader.recursively_compute_variable(
        condensation_rain, DatetimeNoLeap(5, 1, 1),
        DatetimeNoLeap(5, 12, 31),
        intvl_in='monthly')
    filepath = os.path.join(os.path.split(ROOT_PATH)[0], 'netcdf',
                            '00050101.precip_monthly.nc')
    expected = _open_ds_catch_warnings(filepath)['condensation_rain']
    np.testing.assert_array_equal(result.values, expected.values)


def test_recursively_compute_variable_one_level(load_variable_data_loader):
    one_level = Var(
        name='one_level', variables=(condensation_rain, condensation_rain),
        func=lambda x, y: x + y)
    result = load_variable_data_loader.recursively_compute_variable(
        one_level, DatetimeNoLeap(5, 1, 1), DatetimeNoLeap(5, 12, 31),
        intvl_in='monthly')
    filepath = os.path.join(os.path.split(ROOT_PATH)[0], 'netcdf',
                            '00050101.precip_monthly.nc')
    expected = 2. * _open_ds_catch_warnings(filepath)['condensation_rain']
    np.testing.assert_array_equal(result.values, expected.values)


def test_recursively_compute_variable_multi_level(load_variable_data_loader):
    one_level = Var(
        name='one_level', variables=(condensation_rain, condensation_rain),
        func=lambda x, y: x + y)
    multi_level = Var(
        name='multi_level', variables=(one_level, condensation_rain),
        func=lambda x, y: x + y)
    result = load_variable_data_loader.recursively_compute_variable(
        multi_level, DatetimeNoLeap(5, 1, 1), DatetimeNoLeap(5, 12, 31),
        intvl_in='monthly')
    filepath = os.path.join(os.path.split(ROOT_PATH)[0], 'netcdf',
                            '00050101.precip_monthly.nc')
    expected = 3. * _open_ds_catch_warnings(filepath)['condensation_rain']
    np.testing.assert_array_equal(result.values, expected.values)


def test_recursively_compute_grid_attr(load_variable_data_loader):
    result = load_variable_data_loader.recursively_compute_variable(
        bk, DatetimeNoLeap(5, 1, 1),
        DatetimeNoLeap(5, 12, 31), model=example_model,
        intvl_in='monthly')
    filepath = os.path.join(os.path.split(ROOT_PATH)[0], 'netcdf',
                            '00060101.sphum_monthly.nc')
    expected = _open_ds_catch_warnings(filepath)['bk']
    np.testing.assert_array_equal(result.values, expected.values)


def test_recursively_compute_grid_attr_multi_level(load_variable_data_loader):
    one_level = Var(
        name='one_level', variables=(bk, ),
        func=lambda x: 2 * x)
    multi_level = Var(
        name='multi_level', variables=(one_level, bk),
        func=lambda x, y: x + y)
    result = load_variable_data_loader.recursively_compute_variable(
        multi_level, DatetimeNoLeap(5, 1, 1),
        DatetimeNoLeap(5, 12, 31), model=example_model,
        intvl_in='monthly')
    filepath = os.path.join(os.path.split(ROOT_PATH)[0], 'netcdf',
                            '00060101.sphum_monthly.nc')
    expected = 3 * _open_ds_catch_warnings(filepath)['bk']
    np.testing.assert_array_equal(result.values, expected.values)


def test_recursively_compute_grid_attr_error(load_variable_data_loader):
    # Should fail because zsurf is not provided to the example_model object
    zsurf = Var(name=ZSURF_STR, def_time=False, def_vert=False,
                def_lon=True, def_lat=True)
    with pytest.raises(AttributeError):
        load_variable_data_loader.recursively_compute_variable(
            zsurf, DatetimeNoLeap(5, 1, 1),
            DatetimeNoLeap(5, 12, 31), model=example_model,
            intvl_in='monthly')


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python
"""Main script for executing calculations using aospy.

Before using this script
------------------------

It is best to copy this template into a separate directory before populating it
with the objects from your own object library.  You can also always get a fresh
copy from https://github.com/spencerahill/aospy/examples/aospy_main.py

How to use this script
----------------------

On the example library and data
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This script comes pre-populated with objects taken from the example aospy
object library that is included in this directory in the `example_obj_lib.py`
module. So you can try it out on the sample data without modifying anything at
all.

This simple example library includes only one Proj, one Model, and one Run, but
it also includes multiple Var and Region objects over which you can automate
computations.  The date range, sub-annual averaging period, and output data
types can all also be modified.

On your own library and data
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

As a user, there are only two places you need to modify the code:

1. (Only done once) Replace `example_obj_lib` with the name of your object
   library.  Consult the documentation for instructions on how to make your
   object library findable by Python if this statement generates errors.

2. (Done as often as desired) Replace the dummy project, model, run, var, and
   region objects with your objects that you want to perform
   calculations with.

   Also alter the other parameters -- date_range, etc. -- to your liking.

Running the script
------------------

Once the parameters are all set as desired, execute the script from the command
line ::

  ./aospy_main.py  # after `cd` to the directory where you've made your copy

"""
# C.f. instructions above, replace `example_obj_lib` with your own object
# library if you wish to use this on your own data.
from aospy import submit_mult_calcs
import example_obj_lib as lib


# This dictionary contains all of the specifications of calculations that you
# wish to permute over.
calc_suite_specs = dict(
    # Consult `Calc` API reference for further explanation of each option and
    # accepted values.

    # The desired library of aospy objects.
    library=lib,
    # List of the Proj objects, or 'default', or 'all'.
    projects=[lib.example_proj],
    # List of the Model objects, or 'default', or 'all'.
    models=[lib.example_model],
    # List of the Run objects, or 'default', or 'all'.
    runs=[lib.example_run],
    # List of the Var objects, or 'default', or 'all'.
    variables=[lib.precip_largescale, lib.precip_convective, lib.precip_total,
               lib.precip_conv_frac],
    # List of the Region objects, or 'default', or 'all'.
    regions='all',

    # Start and end dates (inclusive).  Either 'default' or a list comprising
    # tuples of the form (start_date, end_date), where start_date and end_date
    # are datetime.datetime objects.  Be sure to add `import datetime` above if
    # using `datetime.datetime` objects.
    date_ranges='default',

    # Sub-annual time-interval to average over.  List of 'ann', seasonal
    # string (e.g. 'djf'), or month integer (1 for Jan, 2 for Feb, etc).
    output_time_intervals=['ann'],
    # List of strings indicating the desired spatiotemporal reductions.
    output_time_regional_reductions=['av', 'std', 'reg.av', 'reg.std',
                                     'reg.ts'],
    # List of desired vertical reductions to perform.
    output_vertical_reductions=[None],

    # List of time spacing of input data.
    input_time_intervals=['monthly'],
    # List of time type of input data.
    input_time_datatypes=['ts'],
    # List the time offset dictionaries (if desired) to apply to the input
    # data (e.g. [{'days': -15}]).
    input_time_offsets=[None],
    # List of vertical data type of input data.
    input_vertical_datatypes=[False],

)


# This dictionary contains options regarding how the calculations are displayed
# to you, submitted for execution, and saved upon execution.
calc_exec_options = dict(
    # List calculations to be performed and prompt for your verification before
    # submitting them for execution.
    prompt_verify=True,

    # Submit all calculations in parallel.  If parallelize is True and client
    # is None, a LocalCluster will be started; the client argument can be used
    # to specify an external dask.distributed Client for use in parallelizing
    # computations
    parallelize=False,
    client=None,

    # Save results of calculations to .tar files, one for each Run object.
    # These tar files are placed using the same directory structure as the
    # standard output relative to their root directory, which is specified via
    # the `tar_direc_out` argument of each Proj object's instantiation.
    write_to_tar=True,
)


# Don't modify this statement.
if __name__ == '__main__':
    calcs = submit_mult_calcs(calc_suite_specs, calc_exec_options)



import os
import sys
import warnings

import pytest

import aospy


ON_WINDOWS = sys.platform == 'win32' or sys.platform == 'win64'


@pytest.mark.skipif(
    ON_WINDOWS, reason='This method of running Jupyter '
    'notebooks on AppVeyor no longer seems to work.')
def test_tutorial_notebook():
    pytest.importorskip('nbformat')
    pytest.importorskip('nbconvert')
    pytest.importorskip('matplotlib')

    import nbformat
    from nbconvert.preprocessors import ExecutePreprocessor

    rootdir = os.path.join(aospy.__path__[0], 'examples')
    with open(os.path.join(rootdir, 'tutorial.ipynb')) as nb_file:
        notebook = nbformat.read(nb_file, as_version=nbformat.NO_CONVERT)
    kernel_name = 'python' + str(sys.version[0])
    ep = ExecutePreprocessor(kernel_name=kernel_name)
    with warnings.catch_warnings(record=True):
        ep.preprocess(notebook, {})
